from pkg_resources import resource_string
import json


def get_asset(filename):
    return resource_string('ftw.crawler.tests.assets', filename)


class MockRequest(object):

    def __init__(self, url='http://example.org'):
        self.url = url


class MockResponse(object):

    def __init__(self, content=None, status_code=200, headers=None,
                 request=None, is_redirect=False):
        self.content = content
        self.status_code = status_code

        if headers is None:
            headers = {}
        self.headers = headers

        if request is None:
            request = MockRequest()
        self.request = request

        self.is_redirect = is_redirect

    @property
    def text(self):
        return self.content.decode('utf-8')

    def json(self):
        return json.loads(self.content)


class MockConverter(object):

    def __init__(self, metadata=None, text=''):
        if metadata is None:
            metadata = {}
        self.metadata = metadata

        self.text = text

    def extract_metadata(self, resource_info):
        return self.metadata

    def extract_text(self, resource_info):
        return self.text



from argparse import Namespace
from ftw.crawler.configuration import Config
from ftw.crawler.configuration import Field
from ftw.crawler.configuration import get_config
from ftw.crawler.configuration import Site
from ftw.crawler.exceptions import NoSuchField
from ftw.crawler.extractors import Extractor
from ftw.crawler.testing import CrawlerTestCase
from pkg_resources import resource_filename


BASIC_CONFIG = resource_filename('ftw.crawler.tests.assets', 'basic_config.py')


class TestConfig(CrawlerTestCase):

    def setUp(self):
        CrawlerTestCase.setUp(self)
        self.site = Site('http://example.org')
        self.tika = 'http://localhost:9998'
        self.solr = 'http://localhost:8983/solr'
        self.slacktoken = 'token'
        self.slackchannel = '#channel'
        self.unique_field = 'UID'
        self.url_field = 'url'
        self.last_modified_field = 'modified'
        self.field = Field('foo', extractor=Extractor())

        self.config = Config([self.site], self.unique_field, self.url_field,
                             self.last_modified_field, [self.field],
                             self.tika, self.solr,
                             self.slacktoken, self.slackchannel)

    def test_config_stores_sites(self):
        self.assertEquals([self.site], self.config.sites)

    def test_config_stores_tika(self):
        self.assertEquals(self.tika, self.config.tika)

    def test_config_stores_solr(self):
        self.assertEquals(self.solr, self.config.solr)

    def test_config_stores_slacktoken(self):
        self.assertEquals(self.slacktoken, self.config.slacktoken)

    def test_config_stores_slackchannel(self):
        self.assertEquals(self.slackchannel, self.config.slackchannel)

    def test_config_stores_unique_field(self):
        self.assertEquals(self.unique_field, self.config.unique_field)

    def test_config_stores_url_field(self):
        self.assertEquals(self.url_field, self.config.url_field)

    def test_config_stores_last_modified_field(self):
        self.assertEquals(
            self.last_modified_field, self.config.last_modified_field)

    def test_config_stores_fields(self):
        self.assertEquals([self.field], self.config.fields)

    def test_binds_fields_to_self(self):
        self.assertEquals(self.field.config, self.config)

    def test_binds_sites_to_self(self):
        self.assertEquals(self.site.config, self.config)

    def test_get_field_retrieves_field_by_name(self):
        field = self.config.get_field('foo')
        self.assertEquals(field, self.field)

    def test_get_field_raises_if_field_not_found(self):
        with self.assertRaises(NoSuchField):
            self.config.get_field('doesnt_exist')


class TestSite(CrawlerTestCase):

    def test_site_requires_url(self):
        with self.assertRaises(TypeError):
            Site()

    def test_site_stores_url(self):
        url = 'http://example.org'
        site = Site(url)
        self.assertEquals(url, site.url)

    def test_site_stores_attributes(self):
        url = 'http://example.org'
        attributes = {'name': 'My Site'}
        site = Site(url, attributes=attributes)
        self.assertEquals({'name': 'My Site'}, site.attributes)


class TestField(CrawlerTestCase):

    def setUp(self):
        CrawlerTestCase.setUp(self)
        self.name = 'Title'
        self.extractor = Extractor()
        self.type_ = str

    def test_field_stores_name(self):
        field = Field(self.name, self.extractor)
        self.assertEquals(self.name, field.name)

    def test_field_stores_extractor(self):
        field = Field(self.name, self.extractor)
        self.assertEquals(self.extractor, field.extractor)

    def test_field_stores_type(self):
        field = Field(self.name, self.extractor, self.type_)
        self.assertEquals(self.type_, field.type_)

    def test_field_stores_required(self):
        field = Field(self.name, self.extractor, self.type_, required=True)
        self.assertEquals(True, field.required)

    def test_field_binds_extractor_to_self(self):
        extractor = Extractor()
        field = Field(self.name, extractor, self.type_, required=True)
        self.assertEquals(field, extractor.field)


class TestGetConfig(CrawlerTestCase):

    def test_get_config_loads_config_module_and_returns_config_instance(self):
        options = Namespace(tika=None, solr=None,
                            slacktoken=None, slackchannel=None)
        options.config = BASIC_CONFIG

        config = get_config(options)
        self.assertIsInstance(config, Config)

    def test_get_config_sets_arguments_from_command_line(self):
        options = Namespace(tika='http://tika', solr='http://solr',
                            slacktoken='token', slackchannel='#channel')
        options.config = BASIC_CONFIG

        config = get_config(options)
        self.assertEquals('http://tika', config.tika)
        self.assertEquals('http://solr', config.solr)
        self.assertEquals('token', config.slacktoken)
        self.assertEquals('#channel', config.slackchannel)



from lxml import etree
from pkg_resources import resource_stream


XML_TYPES = ['application/xml', 'application/xhtml+xml', 'text/xml']
HTML_TYPES = ['text/html']
MARKUP_TYPES = XML_TYPES + HTML_TYPES


def remove_namespaces(tree):
    xslt_file = resource_stream(
        'ftw.crawler.xml_utils', 'remove_namespaces.xsl')
    xslt = etree.parse(xslt_file)
    transform = etree.XSLT(xslt)
    try:
        ns_free_tree = transform(tree)
    except etree.XSLTApplyError:
        return tree
    return etree.ElementTree(ns_free_tree.getroot())



from datetime import datetime
from ftw.crawler.testing import CrawlerTestCase
from ftw.crawler.utils import ExtendedJSONEncoder
from ftw.crawler.utils import from_http_datetime
from ftw.crawler.utils import from_iso_datetime
from ftw.crawler.utils import get_content_type
from ftw.crawler.utils import normalize_whitespace
from ftw.crawler.utils import to_http_datetime
from ftw.crawler.utils import to_iso_datetime
from ftw.crawler.utils import to_utc
from pytz import timezone
import pytz


class TestGetContentType(CrawlerTestCase):

    def test_strips_charset(self):
        header_value = 'text/html; charset=utf-8'
        self.assertEquals('text/html', get_content_type(header_value))

    def test_also_works_for_mimetype_only(self):
        header_value = 'text/html'
        self.assertEquals('text/html', get_content_type(header_value))


class TestToUTC(CrawlerTestCase):

    def test_doesnt_change_dt_already_in_utc(self):
        dt = datetime(2014, 12, 31, 15, 45, 30, tzinfo=pytz.utc)
        self.assertEquals(dt, to_utc(dt))

    def test_assigns_utc_to_naive_dt(self):
        naive_dt = datetime(2014, 12, 31, 15, 45, 30)
        self.assertEquals(datetime(2014, 12, 31, 15, 45, 30, tzinfo=pytz.utc),
                          to_utc(naive_dt))

    def test_converts_tz_aware_dt_to_utc(self):
        cet = timezone('CET')
        dt = datetime(2014, 12, 31, 16, 45, 30, tzinfo=cet)
        self.assertEquals(datetime(2014, 12, 31, 15, 45, 30, tzinfo=pytz.utc),
                          to_utc(dt))


class TestToISODateTime(CrawlerTestCase):

    def test_serializes_naive_datetime_to_utc_iso_8601(self):
        dt = datetime(2014, 12, 31, 15, 45, 30, 999)
        self.assertEquals('2014-12-31T15:45:30.000999Z',
                          to_iso_datetime(dt))

    def test_serializes_tz_aware_datetime_to_utc_iso_8601(self):
        cet = timezone('CET')
        dt = datetime(2014, 12, 31, 15, 45, 30, 999, tzinfo=cet)
        self.assertEquals('2014-12-31T14:45:30.000999Z',
                          to_iso_datetime(dt))


class TestFromISODateTime(CrawlerTestCase):

    def test_parses_utc_iso_8601_to_utc_datetime(self):
        dt = to_utc(datetime(2014, 12, 31, 15, 45, 30))
        self.assertEquals(dt, from_iso_datetime('2014-12-31T15:45:30Z'))
        self.assertEquals(dt, from_iso_datetime('2014-12-31T15:45:30'))
        self.assertEquals(dt, from_iso_datetime('2014-12-31T15:45:30+00:00'))
        self.assertEquals(dt, from_iso_datetime('2014-12-31T16:45:30+01:00'))


class TestToHTTPDateTime(CrawlerTestCase):

    def test_serializes_naive_datetime_to_rfc_2616_http_date(self):
        dt = datetime(2014, 12, 31, 15, 45, 30, 999)
        self.assertEquals('Wed, 31 Dec 2014 15:45:30 GMT',
                          to_http_datetime(dt))

    def test_serializes_tz_aware_datetime_to_rfc_2616_http_date(self):
        cet = timezone('CET')
        dt = datetime(2014, 12, 31, 15, 45, 30, 999, tzinfo=cet)
        self.assertEquals('Wed, 31 Dec 2014 14:45:30 GMT',
                          to_http_datetime(dt))


class TestFromHTTPDateTime(CrawlerTestCase):

    def test_parses_rfc2616_http_date_to_utc_datetime(self):
        from_http = from_http_datetime
        dt = to_utc(datetime(2014, 6, 30, 15, 45, 30))
        # Test parsing of the three formats described by RFC 2616

        # RFC 1123
        self.assertEquals(dt, from_http('Mon, 30 Jun 2014 15:45:30 GMT'))
        # RFC 850
        self.assertEquals(dt, from_http('Monday, 30-Jun-14 15:45:30 GMT'))
        # ANSI C's asctime() Sun Nov  6 08:49:37 1994
        self.assertEquals(dt, from_http('Mon Jun 30 15:45:30 2014'))

        # Test a date in winter as well to trigger any DST issues
        dt_s = to_utc(datetime(2014, 12, 31, 15, 45, 30))
        self.assertEquals(dt_s, from_http('Wed, 31 Dec 2014 15:45:30 GMT'))


class TestNormalizeWhitespace(CrawlerTestCase):

    def test_replaces_tabs_with_space(self):
        self.assertEquals('x x', normalize_whitespace('x\tx'))

    def test_replaces_cr_with_space(self):
        self.assertEquals('x x', normalize_whitespace('x\rx'))

    def test_replaces_lf_with_space(self):
        self.assertEquals('x x', normalize_whitespace('x\nx'))

    def test_strips_leading_and_trailing_whitespace(self):
        self.assertEquals('x', normalize_whitespace(' \r\n \tx \r\n \t '))

    def test_replaces_multiple_whitespaces_by_one(self):
        self.assertEquals('a b c', normalize_whitespace('a  b     c'))


class TestExtendedJSONEncoder(CrawlerTestCase):

    def test_serializes_datetime(self):
        encoder = ExtendedJSONEncoder()
        dt = datetime(2014, 12, 31, 15, 45, 30, 999)
        self.assertEquals('["2014-12-31T15:45:30.000999Z"]',
                          encoder.encode([dt]))

    def test_delegates_unknown_objects_to_default_encoder(self):
        encoder = ExtendedJSONEncoder()
        data = [object()]
        with self.assertRaises(TypeError):
            encoder.encode(data)



from ftw.crawler.log import setup_logging
import argparse
import sys


setup_logging()


def parse_args(argv=None):
    if argv is None:
        argv = sys.argv[1:]

    parser = argparse.ArgumentParser()
    parser.add_argument('config', help='Path to the config file')
    parser.add_argument('url', help='If given, only index the supplied URL',
                        nargs='?', default=None)
    parser.add_argument('--tika', help='Base URL to Tika', metavar='TIKA_URL')
    parser.add_argument('--solr', help='Base URL to Solr', metavar='SOLR_URL')
    parser.add_argument('--slacktoken', help='Token for Slack messages',
                        metavar='SLACK_TOKEN')
    parser.add_argument('--slackchannel', help='Channel for Slack messages',
                        metavar='SLACK_CHANNEL')
    parser.add_argument('-f', '--force', help="Force crawling even if"
                        "document hasn't been modified", action='store_true')
    args = parser.parse_args(argv)

    return args



from argparse import Namespace
from ftw.crawler.fetcher import ResourceFetcher
from ftw.crawler.resource import ResourceInfo
from ftw.crawler.sitemap import Sitemap
from ftw.crawler.tests.helpers import get_asset
from ftw.crawler.tests.helpers import MockResponse
from lxml import etree
from unittest2 import TestCase
import calendar
import json
import logging
import requests


def timestamp(dt):
    return calendar.timegm(dt.utctimetuple())


class CrawlerTestCase(TestCase):

    def setUp(self):
        logging.disable(logging.CRITICAL)

    def tearDown(self):
        logging.disable(logging.NOTSET)


class XMLTestCase(CrawlerTestCase):

    def tostring(self, xml):
        return etree.tostring(xml, xml_declaration=True,
                              encoding='utf-8', pretty_print=True)

    def tree_from_str(self, xml):
        parser = etree.XMLParser(remove_blank_text=True)
        root = etree.XML(xml.strip(), parser=parser)
        return root.getroottree()

    def assertXMLEquals(self, expected, actual):
        """Assert two strings of XML are equal by normalizing
        whitespace outside text nodes.
        """
        actual_xml = self.tostring(self.tree_from_str(actual))
        expected_xml = self.tostring(self.tree_from_str(expected))
        return self.assertEquals(expected_xml, actual_xml)


class DatetimeTestCase(CrawlerTestCase):

    def assertDatetimesAlmostEqual(self, expected, actual, delta=2):
        expected_ts = timestamp(expected)
        actual_ts = timestamp(actual)
        msg = "datetimes {} and {} aren't within {} seconds.".format(
            expected, actual, delta)
        return self.assertAlmostEqual(expected_ts, actual_ts,
                                      delta=delta, msg=msg)


class FetcherTestCase(CrawlerTestCase):

    def _create_fetcher(self, resource_info=None, session=None, tempdir=None,
                        options=None):
        if resource_info is None:
            resource_info = ResourceInfo()

        if session is None:
            session = requests.Session()

        if options is None:
            options = Namespace(force=False)

        return ResourceFetcher(
            resource_info=resource_info, session=session, tempdir=tempdir,
            options=options)


SITEMAP = get_asset('sitemap.xml')


class SitemapTestCase(CrawlerTestCase):

    def create_sitemap(self, urls=None, site=None, sitemap_url=None):
        sitemap = Sitemap(site, SITEMAP, url=sitemap_url)
        del sitemap.tree
        url_infos = [{'loc': url} for url in urls]
        sitemap._url_infos = url_infos
        return sitemap


SOLR_RESULTS_TEMPLATE = """\
{
  "responseHeader": {
    "status": 0,
    "QTime": 0,
    "params": {
    }
  },
  "response": {
    "numFound": %s,
    "start": 0,
    "docs": %s
  }
}
"""


class SolrTestCase(CrawlerTestCase):

    def create_solr_response(self, status=200, content=None, headers=None):
        if content is None:
            content = '{"responseHeader":{"status":%s,"QTime":10}}' % status

        if headers is None:
            headers = {'content-type': 'application/json; charset=UTF-8'}

        response = MockResponse(
            status_code=status,
            content=content,
            headers={'content-type': 'application/json; charset=UTF-8'})
        return response

    def create_solr_results(self, docs):
        docs = json.dumps(docs)
        content = SOLR_RESULTS_TEMPLATE % (len(docs), docs)
        response = self.create_solr_response(content=content)
        return response



from datetime import datetime
from ftw.crawler.configuration import Config
from ftw.crawler.configuration import Field
from ftw.crawler.configuration import Site
from ftw.crawler.extractors import ConstantExtractor
from ftw.crawler.extractors import CreatorExtractor
from ftw.crawler.extractors import DescriptionExtractor
from ftw.crawler.extractors import FieldMappingExtractor
from ftw.crawler.extractors import HeaderMappingExtractor
from ftw.crawler.extractors import IndexingTimeExtractor
from ftw.crawler.extractors import KeywordsExtractor
from ftw.crawler.extractors import LastModifiedExtractor
from ftw.crawler.extractors import PlainTextExtractor
from ftw.crawler.extractors import SiteAttributeExtractor
from ftw.crawler.extractors import SlugExtractor
from ftw.crawler.extractors import SnippetTextExtractor
from ftw.crawler.extractors import TargetURLExtractor
from ftw.crawler.extractors import TitleExtractor
from ftw.crawler.extractors import UIDExtractor
from ftw.crawler.extractors import URLExtractor


PORTAL_TYPE_MAPPING = {
    'text/html': 'ContentPage',
    'application/pdf': 'File',
}

OBJECT_TYPE_MAPPING = {
    'ContentPage': 'CONTENT_PAGE',
    'File': 'FILE',
}


# May be overriden via command line arguments
TIKA_URL = 'http://localhost:9998/'
SOLR_URL = 'http://localhost:8983/solr'
SLACK_TOKEN = 'this-is-a-slack-token'
SLACK_CHANNEL = '#slack-channel'


CONFIG = Config(
    sites=[
        Site('https://www.sportamt-bern.ch/',
             attributes={'site_area': 'Sportamt Bern'}),

        Site('http://www.sitemapxml.co.uk/',
             attributes={'site_area': 'Sitemap XML'}),

        Site('http://www.pctipp.ch/',
             attributes={'site_area': 'PCtipp'}),

        Site('http://mailchimp.com',
             attributes={'site_area': 'MailChimp'}),

        Site('https://bgs.zg.ch',
             attributes={'site_area': 'Gesetzessammlung'}),
    ],
    unique_field='UID',
    url_field='path_string',
    last_modified_field='modified',

    fields=[
        Field('allowedRolesAndUsers',
              extractor=ConstantExtractor(['Anonymous']),
              multivalued=True),
        Field('created',
              extractor=LastModifiedExtractor(),
              type_=datetime),
        Field('Creator',
              extractor=CreatorExtractor()),
        Field('Description',
              extractor=DescriptionExtractor()),
        Field('effective',
              extractor=IndexingTimeExtractor(),
              type_=datetime),
        Field('expires',
              extractor=ConstantExtractor(datetime(2050, 12, 31)),
              type_=datetime),
        Field('getId',
              extractor=SlugExtractor()),
        Field('getRemoteUrl',
              extractor=TargetURLExtractor()),
        Field('modified',
              extractor=LastModifiedExtractor(),
              type_=datetime),
        Field('object_type',
              extractor=FieldMappingExtractor(
                  'portal_type', OBJECT_TYPE_MAPPING, default='File')),
        Field('path_string',
              extractor=URLExtractor()),
        Field('portal_type',
              extractor=HeaderMappingExtractor(
                  'content-type', PORTAL_TYPE_MAPPING, default='File')),
        Field('SearchableText',
              extractor=PlainTextExtractor()),
        Field('showinsearch',
              extractor=ConstantExtractor(True),
              type_=bool),
        Field('site_area',
              extractor=SiteAttributeExtractor('site_area'),
              multivalued=True),
        Field('snippetText',
              extractor=SnippetTextExtractor()),
        Field('Subject',
              extractor=KeywordsExtractor(),
              multivalued=True),
        Field('Title',
              extractor=TitleExtractor()),
        Field('topics',
              extractor=KeywordsExtractor(),
              multivalued=True),
        Field('UID',
              extractor=UIDExtractor(),
              required=True),
    ],
    tika=TIKA_URL,
    solr=SOLR_URL,
    slacktoken=SLACK_TOKEN,
    slackchannel=SLACK_CHANNEL,
)



from ftw.crawler.exceptions import NoSitemapFound
from ftw.crawler.utils import gunzip
from ftw.crawler.utils import is_gzipped
from ftw.crawler.xml_utils import remove_namespaces
from lxml import etree
from urlparse import urljoin
import io
import logging
import requests


SITEMAP_INDEX_NAMES = ('', 'sitemap_index.xml', 'sitemap_index.xml.gz')
SITEMAP_NAMES = ('', 'sitemap.xml', 'sitemap.xml.gz')
SITEMAP_NS = 'http://www.sitemaps.org/schemas/sitemap/0.9'
PROPERTIES = ('loc', 'lastmod', 'changefreq', 'priority', 'target')

log = logging.getLogger(__name__)


class SitemapIndexFetcher(object):
    """Looks for a sitemap index on a given site, then downloads it and
    returns a ``SitemapIndex`` object.
    """
    def __init__(self, site):
        self.site = site

    def fetch(self):
        """Discovers and downloads the sitemap index for the given site,
        returning a ``SitemapIndex`` object.
        """
        log.info(u'Fetching sitemap index for {}'.format(self.site.url))
        for sm_idx_name in SITEMAP_INDEX_NAMES:
            url = urljoin(self.site.url, sm_idx_name)
            response = requests.get(url, allow_redirects=False)

            if response.status_code == 200:
                sitemap_idx_xml = response.content

                if is_gzipped(response):
                    sitemap_idx_xml = gunzip(sitemap_idx_xml)

                index = SitemapIndex(self.site, sitemap_idx_xml, url)
                if index.is_sitemap_index():
                    return index

        # No sitemap index found - build a virtual one with a single sitemap
        sitemap = SitemapFetcher(self.site).fetch()
        return VirtualSitemapIndex(self.site, sitemaps=[sitemap])


class SitemapIndex(object):
    """Represents a sitemap index on a site, containing one or more sitemaps.
    """

    def __init__(self, site, sitemap_idx_xml, url=None):
        self.site = site
        self.url = url
        self.tree = self._parse_sitemap_idx_xml(sitemap_idx_xml)

        self._sitemap_infos = None
        self._sitemaps = None

    def is_sitemap_index(self):
        return len(self.tree.xpath('//sitemapindex')) > 0

    @property
    def sitemaps(self):
        """Memoized property that returns a list of ``Sitemap``s contained in
        this sitemap index.
        """
        if self._sitemaps is None:
            self._sitemaps = list(self._fetch_sitemaps())
        return self._sitemaps

    def __contains__(self, url):
        """Tests whether an URL is listed in any of the sitemaps contained in
        this sitemap index (case-insensitive).
        """
        return any(url in sm for sm in self.sitemaps)

    @property
    def sitemap_infos(self):
        """Memoized property that returns a list of dictionaries containing
        the parsed <sitemap> entries.
        """
        if self._sitemap_infos is None:
            self._sitemap_infos = list(self._get_sitemap_infos())
        return self._sitemap_infos

    def _fetch_sitemaps(self):
        for sitemap_info in self.sitemap_infos:
            url = sitemap_info['loc']
            sitemap = SitemapFetcher(self.site).fetch(url)
            yield sitemap

    def _parse_sitemap_idx_xml(self, sitemap_idx_xml):
        tree = etree.parse(io.BytesIO(sitemap_idx_xml))
        tree = remove_namespaces(tree)
        return tree

    def _get_sitemap_infos(self):
        sitemap_nodes = self.tree.xpath('/sitemapindex/sitemap')
        for node in sitemap_nodes:
            sitemap_info = {}
            for name in PROPERTIES:
                results = node.xpath("{}/text()".format(name))
                if results:
                    sitemap_info[name] = results[0]
            yield sitemap_info


class VirtualSitemapIndex(SitemapIndex):
    """A virtual sitemap index - one that is not actually provided by the
    site, but we build ourselves, containing a list of discovered sitemaps.

    The purpose of this is to use the same interface everywhere, even if we're
    just dealing with a single sitemap.
    """

    def __init__(self, site, sitemaps, url=None):
        self.site = site
        self._sitemaps = sitemaps
        self.url = url

    @property
    def sitemap_infos(self):
        raise NotImplementedError


class SitemapFetcher(object):
    """Looks for a sitemap on a given site, then downloads it and returns a
    ``Sitemap`` object.
    """
    def __init__(self, site):
        self.site = site

    def fetch(self, url=None):
        """Discovers and downloads the sitemap for the given site, returning
        a ``Sitemap`` object.
        """
        if url is not None:
            # We're given an URL to a sitemap, don't do any discovery
            log.info(u'Fetching sitemap {}'.format(url))
            response = requests.get(url)
            sitemap_xml = response.content
            if is_gzipped(response):
                sitemap_xml = gunzip(sitemap_xml)
            return Sitemap(self.site, sitemap_xml, url)

        # No URL given, look for sitemap in common locations
        log.info(u'Fetching sitemap for {}'.format(self.site.url))
        for sm_name in SITEMAP_NAMES:
            url = urljoin(self.site.url, sm_name)
            response = requests.get(url)

            if response.status_code == 200:
                sitemap_xml = response.content

                if is_gzipped(response):
                    sitemap_xml = gunzip(sitemap_xml)

                sitemap = Sitemap(self.site, sitemap_xml, url)
                if sitemap.is_sitemap():
                    return sitemap

        raise NoSitemapFound(
            "No sitemap found for {}!".format(self.site.url))


class Sitemap(object):
    """Represents a single sitemap on a site.
    """

    def __init__(self, site, sitemap_xml, url=None):
        self.site = site
        self.url = url
        self.tree = self._parse_sitemap_xml(sitemap_xml)
        self._url_infos = None

    def is_sitemap(self):
        return len(self.tree.xpath('//urlset')) > 0

    @property
    def url_infos(self):
        """Memoized property that returns a list of dictionaries containing
        the parsed <urlinfo>s.
        """
        if self._url_infos is None:
            self._url_infos = list(self._get_url_infos())
        return self._url_infos

    def __contains__(self, url):
        """Tests whether an URL is listed in this sitemap (case-insensitive).
        """
        url_infos = self.url_infos
        return url.lower() in (ui['loc'].lower() for ui in url_infos)

    def _parse_sitemap_xml(self, sitemap_xml):
        tree = etree.parse(io.BytesIO(sitemap_xml))
        tree = remove_namespaces(tree)
        return tree

    def _get_url_infos(self):
        url_nodes = self.tree.xpath('/urlset/url')
        for node in url_nodes:
            url_info = {}
            for name in PROPERTIES:
                results = node.xpath("{}/text()".format(name))
                if results:
                    url_info[name] = results[0]
            yield url_info



class FtwCrawlerException(Exception):
    """Base class for all our own exceptions.
    """


class FetchingError(FtwCrawlerException):
    """An error happend while attempting to fetch a resource.
    """


class ExtractionError(FtwCrawlerException):
    """An error happend while attempting to apply an extractor.
    """


class NoValueExtracted(ExtractionError):
    """The extractor was unable to extract a value.
    """


class SolrError(FtwCrawlerException):
    """Solr returned a non-200 response for an operation.
    """


class NoSuchField(FtwCrawlerException):
    """A field that doesn't exist was specified by name.
    """


class AttemptedRedirect(FtwCrawlerException):
    """An URL attempted a redirect to another location.
    """


class NotModified(FtwCrawlerException):
    """A resource hasn't been modified since the last time it got indexed.
    """


class ConfigError(FtwCrawlerException):
    """A configuration error occurred.
    """


class SiteNotFound(ConfigError):
    """The requested site could not be found in the configuration.
    """


class NoSitemapFound(FtwCrawlerException):
    """No sitemap could be found for the given site.
    """



from ftw.crawler.solr import SolrConnector
import logging


log = logging.getLogger(__name__)


def purge_removed_docs_from_index(config, sitemap_index, indexed_docs):
    solr = SolrConnector(config.solr)
    unique_field = config.unique_field
    url_field = config.url_field
    site = sitemap_index.site

    log.info(u'Purging removed docs from index for {}...'.format(site.url))
    docs_to_purge = []

    for doc in indexed_docs:
        url = doc[url_field]
        uid = doc[unique_field]

        url_in_site = url.startswith(site.url)
        url_in_any_sitemap = any(url in sm for sm in sitemap_index.sitemaps)

        if url_in_site and not url_in_any_sitemap:
            docs_to_purge.append((uid, url))

    for uid, url in docs_to_purge:
        log.info(u'Purging document {} ({}) from Solr'.format(uid, url))
        solr.delete(uid)
    log.info(u'Done purging.')



from ftw.crawler.metadata import SimpleMetadata
import csv
import io
import logging
import requests


log = logging.getLogger(__name__)


class TikaConverter(object):

    def __init__(self, tika_url):
        self.tika_url = tika_url.rstrip('/')

    def _tika_request(self, endpoint, resource_info, headers):
        tika_endpoint = '/'.join((self.tika_url, endpoint))
        with open(resource_info.filename) as fileobj:
            response = requests.put(
                tika_endpoint, data=fileobj, headers=headers)
        return response

    def extract_metadata(self, resource_info):
        log.debug(u"Extracting metadata from '{}' with "
                  "tika JAXRS server.".format(resource_info.filename))

        headers = {'Content-type': resource_info.content_type}
        response = self._tika_request('meta', resource_info, headers)

        csv_file = io.BytesIO(response.content)
        csv_reader = csv.reader(csv_file, delimiter=',', quotechar='"')

        metadata = {}
        for item in iter(csv_reader):
            # In rare cases, Tika returns more than one value per key. So we
            # just join everything after the key using a space, which won't
            # change anything for 99% of items that are just key/value pairs.
            key = item[0]
            value = ' '.join(item[1:])
            metadata[key] = value

        # TODO: We assume Tika returns utf-8 encoded metadata
        for key, value in metadata.items():
            metadata[key] = value.decode('utf-8')

        return SimpleMetadata(metadata)

    def extract_text(self, resource_info):
        log.debug(u"Extracting plain text from '{}' with "
                  "tika JAXRS server.".format(resource_info.filename))

        headers = {'Content-type': resource_info.content_type,
                   'Accept': 'text/plain'}
        response = self._tika_request('tika', resource_info, headers)

        # Normally we would just use response.text to get the decoded response
        # body in unicode, but Tika uses UTF-8 *without declaring it*.
        # See https://issues.apache.org/jira/browse/TIKA-912
        return response.content.decode('utf-8')



from ftw.crawler.configuration import Site
from ftw.crawler.exceptions import NoSitemapFound
from ftw.crawler.sitemap import Sitemap
from ftw.crawler.sitemap import SitemapFetcher
from ftw.crawler.sitemap import SitemapIndex
from ftw.crawler.sitemap import SitemapIndexFetcher
from ftw.crawler.sitemap import VirtualSitemapIndex
from ftw.crawler.testing import SitemapTestCase
from ftw.crawler.tests.helpers import get_asset
from ftw.crawler.tests.helpers import MockResponse
from mock import patch


XHTML_DOC = get_asset('xhtml_doc.html')
SITEMAP = get_asset('sitemap.xml')
SITEMAP_GZ = get_asset('sitemap.xml.gz')
SITEMAP_REQ_ONLY = get_asset('sitemap_req_only.xml')

SITEMAP_INDEX = get_asset('sitemap_index.xml')
SITEMAP_INDEX_GZ = get_asset('sitemap_index.xml.gz')
SITEMAP_INDEX_REQ_ONLY = get_asset('sitemap_index_req_only.xml')


class TestSitemapFetcher(SitemapTestCase):

    def setUp(self):
        SitemapTestCase.setUp(self)
        self.site = Site('http://example.org/')
        self.response = MockResponse(SITEMAP)

    @patch('requests.get')
    def test_finds_and_fetches_sitemap(self, request):
        request.return_value = self.response
        sm_fetcher = SitemapFetcher(self.site)
        sitemap = sm_fetcher.fetch()
        self.assertEquals(2, len(sitemap.url_infos))

    @patch('requests.get')
    def test_falls_back_to_gzipped_sitemap(self, request):
        responses = {
            'http://example.org/': MockResponse(status_code=404),
            'http://example.org/sitemap.xml': MockResponse(
                status_code=404),
            'http://example.org/sitemap.xml.gz': MockResponse(
                status_code=200,
                content=SITEMAP_GZ,
                headers={'Content-Type': 'application/x-gzip'})}

        request.side_effect = lambda url: responses[url]
        sm_fetcher = SitemapFetcher(self.site)
        sitemap = sm_fetcher.fetch()

        self.assertEquals(2, len(sitemap.url_infos))

    @patch('requests.get')
    def test_fetches_sitemap_from_url_without_discovery(self, request):
        responses = {
            'http://example.org/sitemap.xml': MockResponse(
                status_code=404),
            'http://example.org/foo/sitemap.xml': MockResponse(
                status_code=200,
                content=SITEMAP),
        }
        request.side_effect = lambda url: responses[url]

        sm_fetcher = SitemapFetcher(self.site)
        sitemap = sm_fetcher.fetch('http://example.org/foo/sitemap.xml')
        self.assertEquals(2, len(sitemap.url_infos))

    @patch('requests.get')
    def test_fetches_gzipped_sitemap_from_url_without_discovery(self, request):
        responses = {
            'http://example.org/sitemap.xml': MockResponse(
                status_code=404),
            'http://example.org/foo/sitemap.xml.gz': MockResponse(
                status_code=200,
                content=SITEMAP_GZ,
                headers={'Content-Type': 'application/x-gzip'})}
        request.side_effect = lambda url: responses[url]

        sm_fetcher = SitemapFetcher(self.site)
        sitemap = sm_fetcher.fetch('http://example.org/foo/sitemap.xml.gz')
        self.assertEquals(2, len(sitemap.url_infos))

    @patch('requests.get')
    def test_raises_if_no_sitemap_found(self, request):
        not_found = MockResponse(status_code=404)
        request.return_value = not_found

        sm_fetcher = SitemapFetcher(self.site)
        with self.assertRaises(NoSitemapFound):
            sm_fetcher.fetch()

    @patch('requests.get')
    def test_decompresses_gzipped_sitemap(self, request):
        not_found = MockResponse(status_code=404)
        html = MockResponse(status_code=200, content=XHTML_DOC)
        found_gz = MockResponse(status_code=200, content=SITEMAP_GZ,
                                headers={'Content-Type': 'application/x-gzip'})
        responses = {'http://example.org/': html,
                     'http://example.org/sitemap.xml': not_found,
                     'http://example.org/sitemap.xml.gz': found_gz}

        request.side_effect = lambda url: responses[url]

        sm_fetcher = SitemapFetcher(self.site)
        sitemap = sm_fetcher.fetch()

        self.assertIn('http://example.org/foo', sitemap)
        self.assertIn('http://example.org/bar', sitemap)
        self.assertEquals(2, len(sitemap.url_infos))

    @patch('requests.get')
    def test_doesnt_choke_on_charset_in_content_type(self, request):
        request.return_value = MockResponse(
            content=SITEMAP,
            headers={'Content-Type': 'text/html; charset=utf-8'})
        sm_fetcher = SitemapFetcher(self.site)
        sitemap = sm_fetcher.fetch()
        self.assertEquals(2, len(sitemap.url_infos))

    @patch('requests.get')
    def test_supports_absolute_sitemap_urls(self, request):
        responses = {
            'http://example.org/foo/bar/seitenkarte': MockResponse(
                status_code=200,
                content=SITEMAP),
        }
        request.side_effect = lambda url: responses[url]

        site = Site('http://example.org/foo/bar/seitenkarte')
        sm_fetcher = SitemapFetcher(site)
        sitemap = sm_fetcher.fetch()
        self.assertEquals(2, len(sitemap.url_infos))


class TestSitemap(SitemapTestCase):

    def setUp(self):
        super(SitemapTestCase, self).setUp()
        self.site = Site('http://example.org/')

    def test_parses_sitemap(self):
        sitemap = Sitemap(self.site, SITEMAP)

        self.assertEqual([
            {'loc': 'http://example.org/foo',
             'changefreq': 'daily',
             'lastmod': '2014-12-31',
             'priority': '1.0'},
            {'loc': 'http://example.org/bar',
             'target': 'http://example.org/target-bar',
             'changefreq': 'daily',
             'lastmod': '2005-01-01',
             'priority': '1.0'}],
            sitemap.url_infos)

    def test_deals_with_missing_elements(self):
        sitemap = Sitemap(self.site, SITEMAP_REQ_ONLY)

        self.assertEqual([
            {'loc': 'http://example.org/foo'},
            {'loc': 'http://example.org/bar'}],
            sitemap.url_infos)

    def test_keeps_reference_to_site(self):
        sitemap = Sitemap(self.site, SITEMAP_REQ_ONLY)

        self.assertEqual(self.site, sitemap.site)

    def test_supports_testing_for_membership(self):
        sitemap = Sitemap(self.site, SITEMAP_REQ_ONLY)

        self.assertIn('http://example.org/foo', sitemap)
        self.assertNotIn('http://example.org/not_contained', sitemap)

    def test_testing_for_membership_is_case_insensitive(self):
        sitemap = Sitemap(self.site, SITEMAP_REQ_ONLY)

        self.assertIn('http://example.org/foo', sitemap)
        self.assertIn('HTTP://EXAMPLE.ORG/FOO', sitemap)


class TestSitemapIndex(SitemapTestCase):

    def setUp(self):
        super(SitemapTestCase, self).setUp()
        self.site = Site('http://example.org/')

    def test_parses_sitemap_index(self):
        sitemap_index = SitemapIndex(self.site, SITEMAP_INDEX)

        self.assertEqual([
            {'loc': 'http://example.org/foo-sitemap.xml',
             'lastmod': '2015-10-23T14:00:05+02:00'},
            {'loc': 'http://example.org/bar-sitemap.xml',
             'lastmod': '2015-10-23T16:00:05+02:00'}],
            sitemap_index.sitemap_infos)

    def test_deals_with_missing_elements(self):
        sitemap_index = SitemapIndex(self.site, SITEMAP_INDEX_REQ_ONLY)

        self.assertEqual([
            {'loc': 'http://example.org/foo-sitemap.xml'},
            {'loc': 'http://example.org/bar-sitemap.xml'}],
            sitemap_index.sitemap_infos)

    def test_keeps_reference_to_site(self):
        sitemap_index = SitemapIndex(self.site, SITEMAP_INDEX)

        self.assertEqual(self.site, sitemap_index.site)

    def test_supports_testing_for_membership(self):
        sitemap_foo = self.create_sitemap(urls=['http://example.org/foo'])
        sitemap_bar = self.create_sitemap(urls=['http://example.org/bar'])
        sitemap_index = VirtualSitemapIndex(
            self.site, sitemaps=[sitemap_foo, sitemap_bar])

        self.assertIn('http://example.org/foo', sitemap_index)
        self.assertIn('http://example.org/bar', sitemap_index)
        self.assertNotIn('http://example.org/not_contained', sitemap_index)

    def test_testing_for_membership_is_case_insensitive(self):
        sitemap = Sitemap(self.site, SITEMAP)
        sitemap_index = VirtualSitemapIndex(self.site, sitemaps=[sitemap])

        self.assertIn('http://example.org/foo', sitemap_index)
        self.assertIn('HTTP://EXAMPLE.ORG/FOO', sitemap_index)


class TestSitemapIndexFetcher(SitemapTestCase):

    def setUp(self):
        SitemapTestCase.setUp(self)
        self.site = Site('http://example.org/')
        self.response = MockResponse(SITEMAP_INDEX)

    @patch('requests.get')
    def test_finds_and_fetches_sitemap_index(self, request):
        request.return_value = self.response
        sm_idx_fetcher = SitemapIndexFetcher(self.site)
        sitemap_index = sm_idx_fetcher.fetch()
        self.assertEquals(2, len(sitemap_index.sitemap_infos))

    @patch('requests.get')
    def test_falls_back_to_gzipped_sitemap_index(self, request):
        responses = {
            'http://example.org/': MockResponse(status_code=404),
            'http://example.org/sitemap_index.xml': MockResponse(
                status_code=404),
            'http://example.org/sitemap_index.xml.gz': MockResponse(
                status_code=200,
                content=SITEMAP_INDEX_GZ,
                headers={'Content-Type': 'application/x-gzip'})}

        request.side_effect = lambda url, **kwargs: responses[url]
        sm_idx_fetcher = SitemapIndexFetcher(self.site)
        sitemap_index = sm_idx_fetcher.fetch()

        self.assertEquals(2, len(sitemap_index.sitemap_infos))

    @patch('requests.get')
    def test_returns_virtual_index_if_no_sitemap_index_found(self, request):
        not_found = MockResponse(status_code=404)
        responses = {
            'http://example.org/sitemap_index.xml': not_found,
            'http://example.org/sitemap_index.xml.gz': not_found,
            'http://example.org/sitemap.xml': MockResponse(
                status_code=200,
                content=SITEMAP),
            'http://example.org/': MockResponse(
                status_code=200,
                content=XHTML_DOC),
        }
        request.side_effect = lambda url, **kwargs: responses[url]

        sm_idx_fetcher = SitemapIndexFetcher(self.site)
        sitemap_index = sm_idx_fetcher.fetch()

        self.assertIsInstance(sitemap_index, VirtualSitemapIndex)
        self.assertIn('http://example.org/foo', sitemap_index)

    @patch('requests.get')
    def test_decompresses_gzipped_sitemap_index(self, request):
        not_found = MockResponse(status_code=404)
        found_gz = MockResponse(status_code=200, content=SITEMAP_INDEX_GZ,
                                headers={'Content-Type': 'application/x-gzip'})
        responses = {'http://example.org/': MockResponse(status_code=404),
                     'http://example.org/sitemap_index.xml': not_found,
                     'http://example.org/sitemap_index.xml.gz': found_gz}

        request.side_effect = lambda url, **kwargs: responses[url]

        sm_idx_fetcher = SitemapIndexFetcher(self.site)
        sitemap_index = sm_idx_fetcher.fetch()

        self.assertEquals(2, len(sitemap_index.sitemap_infos))

    @patch('requests.get')
    def test_doesnt_choke_on_charset_in_content_type(self, request):
        request.return_value = MockResponse(
            content=SITEMAP_INDEX,
            headers={'Content-Type': 'text/html; charset=utf-8'})
        sm_idx_fetcher = SitemapIndexFetcher(self.site)
        sitemap_index = sm_idx_fetcher.fetch()
        self.assertEquals(2, len(sitemap_index.sitemap_infos))

    @patch('requests.get')
    def test_supports_absolute_sitemap_index_urls(self, request):
        responses = {
            'http://example.org/foo/bar/seitenverzeichnisse': MockResponse(
                status_code=200,
                content=SITEMAP_INDEX),
        }
        request.side_effect = lambda url, **kwargs: responses[url]

        site = Site('http://example.org/foo/bar/seitenverzeichnisse')
        sm_idx_fetcher = SitemapIndexFetcher(site)
        sitemap_index = sm_idx_fetcher.fetch()
        self.assertEquals(2, len(sitemap_index.sitemap_infos))



from ftw.crawler.resource import ResourceInfo
from ftw.crawler.testing import CrawlerTestCase
from ftw.crawler.tests.helpers import MockResponse
from ftw.crawler.tika import TikaConverter
from mock import mock_open
from mock import patch
import csv
import io


# TODO: Figure out how to mock open() without using module globals
open_mock = mock_open()


class TestTikaConverter(CrawlerTestCase):

    def _to_csv(self, mapping):
        csv_buffer = io.BytesIO()
        csv_writer = csv.writer(csv_buffer, delimiter=',', quotechar='"')
        for key, value in mapping.items():
            csv_writer.writerow((key, value))
        return csv_buffer.getvalue()

    @patch('requests.put')
    @patch('ftw.crawler.tika.open', open_mock, create=True)
    def test_extracts_metadata(self, request):
        resource_info = ResourceInfo(content_type='application/pdf')
        metadata = {'title': 'Some title',
                    'creator': 'John Doe'}
        request.return_value = MockResponse(content=self._to_csv(metadata))

        tika = TikaConverter('http://localhost:9998')
        extracted_metadata = tika.extract_metadata(resource_info)

        self.assertEquals(metadata, extracted_metadata)

        for value in extracted_metadata.values():
            if isinstance(value, basestring):
                self.assertIsInstance(value, unicode)

        request.assert_called_with(
            'http://localhost:9998/meta',
            headers={'Content-type': 'application/pdf'},
            data=open_mock.return_value)

    @patch('requests.put')
    @patch('ftw.crawler.tika.open', open_mock, create=True)
    def test_extracts_text(self, request):
        resource_info = ResourceInfo(content_type='application/pdf')
        request.return_value = MockResponse(content='foo bar')

        tika = TikaConverter('http://localhost:9998')
        extracted_text = tika.extract_text(resource_info)

        self.assertEquals(u'foo bar', extracted_text)
        self.assertIsInstance(extracted_text, unicode)
        request.assert_called_with(
            'http://localhost:9998/tika',
            headers={'Content-type': 'application/pdf',
                     'Accept': 'text/plain'},
            data=open_mock.return_value)



class ResourceInfo(object):

    def __init__(self, filename=None, content_type=None, site=None,
                 url_info=None, last_indexed=None, headers=None,
                 metadata=None, text=None):
        self.filename = filename
        self.content_type = content_type
        self.site = site
        self.url_info = url_info
        self.last_indexed = last_indexed
        self.headers = headers
        self.metadata = metadata
        self.text = text



METADATA_MAPPING = {
    'title': ['dcterms:title', 'dc:title', 'DC.title', 'title'],
    'created': ['dcterms:created', 'meta:creation-date', 'Creation-Date'],
    'description': ['dcterms:description', 'dc:description', 'DC.description',
                    'description'],
    'keywords': ['dc:subject', 'DC.subject', 'keywords', 'subject'],
    'creator': ['dcterms:creator', 'dc:creator', 'DC.creator', 'creator',
                'author'],
}


class SimpleMetadata(dict):
    """A dict subclass to map prefixed metadata properties from different
    metadata sets to common, canonical keys.

    E.g. the value for key 'dcterms:title' will be mapped to just 'title'.

    Prefixes take precedence in the order they are specified, i.e. the first
    prefix that is found will be mapped to the canonical key and the search is
    then stopped.
    """

    def __init__(self, mapping):
        dict.__init__(self, mapping)

        simple_metadata = {}
        for property_, keys in METADATA_MAPPING.items():
            for possible_key in keys:
                if possible_key in self:
                    simple_metadata[property_] = self[possible_key]
                    break

        self.update(simple_metadata)



from ftw.crawler import parse_args
from ftw.crawler.configuration import get_config
from ftw.crawler.exceptions import AttemptedRedirect
from ftw.crawler.exceptions import FetchingError
from ftw.crawler.exceptions import NotModified
from ftw.crawler.extractors import ExtractionEngine
from ftw.crawler.fetcher import ResourceFetcher
from ftw.crawler.purging import purge_removed_docs_from_index
from ftw.crawler.resource import ResourceInfo
from ftw.crawler.sitemap import SitemapIndexFetcher
from ftw.crawler.solr import SolrConnector
from ftw.crawler.solr import solr_escape
from ftw.crawler.tika import TikaConverter
from ftw.crawler.utils import from_iso_datetime
import logging
import os
import requests
import shutil
import tempfile
import pkg_resources


log = logging.getLogger(__name__)


try:
    pkg_resources.get_distribution('slacker')
except pkg_resources.DistributionNotFound:
    log.info('Not logging errors to slack (slacker not found)')
    HAS_SLACK = False
else:
    from slack import SlackLogger
    HAS_SLACK = True


def display_fields(field_values):
    log.debug(u"")
    log.debug(u"=== EXTRACTED FIELD VALUES ===")
    for key, value in field_values.items():
        if key in ('SearchableText', 'snippetText'):
            value = value.replace('\n', '\\n ').replace('\r', '\\r ')
            value = value[:60] + '...'

        log.debug(u"{:<22} {}".format(unicode(key) + u':', value))

    log.debug(u"")


def get_indexed_docs(config, solr, site):
    query = '{}:{}*'.format(config.url_field, solr_escape(site.url))
    indexed_docs = solr.search(
        query,
        fl=(config.unique_field, config.url_field, config.last_modified_field))
    return indexed_docs


def get_indexing_time(url, indexed_docs, config):
    for doc in indexed_docs:
        if doc[config.url_field] == url:
            isodate = doc[config.last_modified_field]
            return from_iso_datetime(isodate)
    return None


def crawl_and_index(tempdir, config, options):
    solr = SolrConnector(config.solr)

    if HAS_SLACK:
        slacklogger = SlackLogger(options.slacktoken)

    for site in config.sites:
        # Skip non-matching sites if we're only indexing a specific URL
        if options.url and not options.url.startswith(site.url):
            continue

        try:
            crawl_site(tempdir, config, options, solr, site)
        except Exception as ex:
            log.error('Failed to crawl {}'.format(site.url))
            log.error('{}: {}'.format(type(ex).__name__, str(ex.message)))
            log.info('Continuing with next site...')
            if HAS_SLACK:
                slacklogger.logError(ex, site, options.slackchannel)
            continue


def crawl_site(tempdir, config, options, solr, site):
    # Fetch and parse the sitemap index (or build a virtual one)
    sitemap_index = SitemapIndexFetcher(site).fetch()

    log.info(u"Crawling {} [{} sitemap(s)]...".format(
        site.url, len(sitemap_index.sitemaps)))

    # Get all docs indexed in Solr for a particular site
    indexed_docs = get_indexed_docs(config, solr, site)

    # Purge docs that have been removed from sitemap(s) from Solr index
    purge_removed_docs_from_index(config, sitemap_index, indexed_docs)

    # Create a requests session to allow for connection pooling
    fetcher_session = requests.Session()

    for sitemap in sitemap_index.sitemaps:
        total = len(sitemap.url_infos)

        log.info(u"-" * 78)
        log.info(u"Crawling {}...".format(sitemap.url))

        for n, url_info in enumerate(sitemap.url_infos, start=1):
            url = url_info['loc']
            progress = '[{}/{}]'.format(n, total)

            # If we're only indexing a specific URL, skip all others
            if options.url and not url == options.url:
                continue

            log.debug(u"{}: {}".format(url, unicode(url_info)))

            # Get time this document was last indexed
            last_indexed = get_indexing_time(url, indexed_docs, config)

            # Fetch and save resource
            resource_info = ResourceInfo(site=sitemap.site,
                                         url_info=url_info,
                                         last_indexed=last_indexed)
            fetcher = ResourceFetcher(
                resource_info, fetcher_session, tempdir, options)
            try:
                resource_info = fetcher.fetch()
            except NotModified:
                log.info(u"{}   Skipped {} (not modified)".format(progress,
                                                                  url))
                continue
            except AttemptedRedirect:
                continue
            except FetchingError, e:
                log.error(unicode(e))
                continue
            except Exception as ex:
                print ex

            # Extract metadata and plain text
            engine = ExtractionEngine(
                config, resource_info, converter=TikaConverter(config.tika))
            field_values = engine.extract_field_values()
            display_fields(field_values)
            os.unlink(resource_info.filename)

            # Index into Solr
            log.debug(u"Indexing {} into solr.".format(url))
            response = solr.index(field_values)
            if response.status_code == 200:
                log.info(u"{} * Indexed {}".format(progress, url))

    log.info(u"=" * 78)
    log.info(u"")


def main():
    options = parse_args()
    config = get_config(options)

    tempdir = tempfile.mkdtemp(prefix='ftw.crawler_')
    log.debug(u"Using temporary directory {}".format(tempdir))
    try:
        crawl_and_index(tempdir, config, options)
    finally:
        shutil.rmtree(tempdir)


if __name__ == '__main__':
    main()



import json
import pkg_resources


try:
    pkg_resources.get_distribution('slacker')
except pkg_resources.DistributionNotFound:
    # To use this module you have to install slacker
    raise ImportError('Please install the extra "slack" to use slack '
                      'integration')
else:
    from slacker import Slacker


class SlackLogger(object):
    def __init__(self, slacktoken):
        self.slack = Slacker(slacktoken)

    def logError(self, ex, site, channel):
        text = "Error while crawling external site indexes!"

        attdata = self.generateAttdata(ex, site)
        channel = self.checkChannel(channel)

        self.send(text, attdata, channel)

    def checkChannel(self, channel):
        if not channel.startswith('#'):
            channel = '#' + channel
        return channel

    def generateAttdata(self, ex, site):
        return json.dumps(
            [
                {
                    "color": 'danger',
                    "fields": [
                        {
                            "title": "Site",
                            "value": site.url
                        },
                        {
                            "title": "Exception Type",
                            "value": type(ex).__name__
                        },
                        {
                            "title": "Error Message",
                            "value": str(ex.message)
                        }
                    ]
                }
            ]
        )

    def send(self, text, attdata, channel):
        username = self.slack.auth.test().body['user']
        self.slack.chat.post_message(channel,
                                     text,
                                     as_user=username,
                                     link_names=1,
                                     attachments=attdata)



from ftw.crawler.exceptions import AttemptedRedirect
from ftw.crawler.exceptions import FetchingError
from ftw.crawler.exceptions import NotModified
from ftw.crawler.utils import from_iso_datetime
from ftw.crawler.utils import get_content_type
import logging
import tempfile
import time


log = logging.getLogger(__name__)


class ResourceFetcher(object):

    def __init__(self, resource_info, session, tempdir, options):
        self.resource_info = resource_info
        self.url_info = resource_info.url_info
        self.session = session
        self.tempdir = tempdir
        self.options = options

    def _mktmp(self):
        return tempfile.NamedTemporaryFile(dir=self.tempdir, delete=False)

    def is_modified(self):
        if self.resource_info.last_indexed is None:
            return True

        if 'lastmod' in self.url_info:
            last_modified = self.url_info['lastmod']
            last_modified = from_iso_datetime(last_modified)
            return last_modified > self.resource_info.last_indexed

        # No 'lastmod' in urlinfo - fall back to a HEAD request
        response = self.session.head(self.url_info['loc'])
        if 'last-modified' in response.headers:
            # TODO: Use 'closing' context manager in order to avoid
            # blocking connection pool
            last_modified = response.headers['last-modified']
            last_modified = from_iso_datetime(last_modified)
            return last_modified > self.resource_info.last_indexed
        return True

    def fetch(self):
        resource_info = self.resource_info
        url = self.url_info['loc']

        modified = self.is_modified()
        if not self.options.force and not modified:
            raise NotModified

        response = self.session.get(url, allow_redirects=False)
        if response.is_redirect:
            # TODO: With redirects it's unclear which URL to use as the
            # canonical URL - so we don't allow them for now.
            log.warn(u"URL {} attempted a redirect - skipped.".format(url))
            raise AttemptedRedirect(url)

        while response.status_code == 429:
            log.warn(u"429 Too Many Requests, sleeping for {}s".format(
                self.resource_info.site.sleeptime))
            time.sleep(self.resource_info.site.sleeptime)
            response = self.session.get(url, allow_redirects=False)
            if response.status_code == 429:
                self.resource_info.site.sleeptime *= 2

        if not response.status_code == 200:
            raise FetchingError(u"Could not fetch {}. Got status {}".format(
                url, response.status_code))

        with self._mktmp() as resource_file:
            resource_file.write(response.content)

        content_type = get_content_type(response.headers.get('Content-Type'))

        resource_info.filename = resource_file.name
        resource_info.content_type = content_type
        resource_info.headers = response.headers
        resource_info.filename = resource_file.name

        log.debug(u"Resource saved to {}".format(resource_info.filename))
        return resource_info



from ftw.crawler.testing import CrawlerTestCase
from ftw.crawler.slack import SlackLogger
from ftw.crawler.configuration import Site
import json


class TestSlack(CrawlerTestCase):
    def setUp(self):
        CrawlerTestCase.setUp(self)
        self.token = 'this-is-a-slack-token'
        self.slacklogger = SlackLogger(self.token)

    def test_slack_token_required_for_logger_init(self):
        with self.assertRaises(TypeError):
            SlackLogger()

    def test_slack_can_handle_all_channels(self):
        # The channel-name always needs a # infornt of it
        no_hashtag = self.slacklogger.checkChannel('no-hashtag')
        hashtag = self.slacklogger.checkChannel('#hashtag')

        self.assertEquals('#no-hashtag', no_hashtag)
        self.assertEquals('#hashtag', hashtag)

    def test_slack_can_format_json_from_data(self):
        ex = Exception('error message')
        site = Site('http://some-url.com/')

        data = json.loads(self.slacklogger.generateAttdata(ex, site))[0]

        self.assertEquals("http://some-url.com/",
                          data['fields'][0]['value'])

        self.assertEquals("Exception",
                          data['fields'][1]['value'])

        self.assertEquals("error message",
                          data['fields'][2]['value'])



from argparse import Namespace
from copy import deepcopy
from datetime import datetime
from ftw.crawler.configuration import Config
from ftw.crawler.configuration import Field
from ftw.crawler.configuration import get_config
from ftw.crawler.configuration import Site
from ftw.crawler.exceptions import ExtractionError
from ftw.crawler.exceptions import NoSuchField
from ftw.crawler.exceptions import NoValueExtracted
from ftw.crawler.extractors import ConstantExtractor
from ftw.crawler.extractors import CreatorExtractor
from ftw.crawler.extractors import DescriptionExtractor
from ftw.crawler.extractors import ExtractionEngine
from ftw.crawler.extractors import Extractor
from ftw.crawler.extractors import FieldMappingExtractor
from ftw.crawler.extractors import FilenameExtractor
from ftw.crawler.extractors import HeaderMappingExtractor
from ftw.crawler.extractors import HTTPHeaderExtractor
from ftw.crawler.extractors import IndexingTimeExtractor
from ftw.crawler.extractors import KeywordsExtractor
from ftw.crawler.extractors import LastModifiedExtractor
from ftw.crawler.extractors import MetadataExtractor
from ftw.crawler.extractors import PlainTextExtractor
from ftw.crawler.extractors import SiteAttributeExtractor
from ftw.crawler.extractors import SlugExtractor
from ftw.crawler.extractors import SnippetTextExtractor
from ftw.crawler.extractors import TargetURLExtractor
from ftw.crawler.extractors import TextExtractor
from ftw.crawler.extractors import TitleExtractor
from ftw.crawler.extractors import UIDExtractor
from ftw.crawler.extractors import URLExtractor
from ftw.crawler.extractors import URLInfoExtractor
from ftw.crawler.extractors import XPathExtractor
from ftw.crawler.resource import ResourceInfo
from ftw.crawler.testing import CrawlerTestCase
from ftw.crawler.testing import DatetimeTestCase
from ftw.crawler.tests.helpers import MockConverter
from ftw.crawler.utils import safe_unicode
from ftw.crawler.utils import to_utc
from mock import MagicMock
from pkg_resources import resource_filename


BASIC_CONFIG = resource_filename('ftw.crawler.tests.assets', 'basic_config.py')


class ExampleMetadataExtractor(MetadataExtractor):

    def extract_value(self, resource_info):
        value = resource_info.metadata.get('example')
        if value is None:
            raise NoValueExtracted
        return safe_unicode(value)


class ExampleTextExtractor(TextExtractor):

    def extract_value(self, resource_info):
        return resource_info.text


class ExampleURLInfoExtractor(URLInfoExtractor):

    def extract_value(self, resource_info):
        return safe_unicode(resource_info.url_info['loc'])


class ExampleHTTPHeaderExtractor(HTTPHeaderExtractor):

    def __init__(self, header_name):
        self.header_name = header_name

    def extract_value(self, resource_info):
        header_value = resource_info.headers[self.header_name]
        # In the rare case of non-ASCII characters in HTTP header field values,
        # we can assume them to be in ISO-8859-1 (see RFC 7230, section 3.2.4)
        return header_value.decode('latin1')


class TestExtractionEngine(CrawlerTestCase):

    def setUp(self):
        CrawlerTestCase.setUp(self)
        args = Namespace(tika=None, solr=None,
                         slacktoken=None, slackchannel=None)
        args.config = BASIC_CONFIG
        self.config = deepcopy(get_config(args))

    def _create_engine(self, config=None, fields=None, resource_info=None,
                       converter=None):
        if config is None:
            config = self.config

        if fields is not None:
            self.config.fields = fields

        if resource_info is None:
            resource_info = ResourceInfo()

        if converter is None:
            converter = MagicMock()

        engine = ExtractionEngine(
            config, resource_info=resource_info, converter=converter)
        return engine

    def test_applies_metadata_extractors_to_converter_metadata(self):
        converter = MockConverter({'example': u'value', 'other': u'data'})
        field = Field('EXAMPLE', extractor=ExampleMetadataExtractor())
        engine = self._create_engine(fields=[field], converter=converter)

        self.assertEquals({'EXAMPLE': u'value'}, engine.extract_field_values())

    def test_applies_text_extractors_to_converter_plain_text(self):
        converter = MockConverter(text=u'foo bar')
        field = Field('EXAMPLE', extractor=ExampleTextExtractor())
        engine = self._create_engine(fields=[field], converter=converter)

        self.assertEquals({'EXAMPLE': u'foo bar'},
                          engine.extract_field_values())

    def test_applies_urlinfo_extractors_to_urlinfo(self):
        field = Field('EXAMPLE', extractor=ExampleURLInfoExtractor())
        resource_info = ResourceInfo(url_info={'loc': 'http://example.org'})
        engine = self._create_engine(
            resource_info=resource_info, fields=[field])

        self.assertEquals({'EXAMPLE': u'http://example.org'},
                          engine.extract_field_values())

    def test_applies_site_config_extractors_to_site(self):
        field = Field('EXAMPLE', extractor=SiteAttributeExtractor('name'))
        site = Site('http://example.org', attributes={'name': 'My Site'})
        resource_info = ResourceInfo(site=site)
        engine = self._create_engine(
            fields=[field], resource_info=resource_info)

        self.assertEquals({'EXAMPLE': u'My Site'},
                          engine.extract_field_values())

    def test_applies_http_header_extractors_to_headers(self):
        field = Field(
            'EXAMPLE', extractor=ExampleHTTPHeaderExtractor('example-header'))
        resource_info = ResourceInfo(headers={'example-header': 'value'})
        engine = self._create_engine(
            fields=[field], resource_info=resource_info)

        self.assertEquals({'EXAMPLE': u'value'},
                          engine.extract_field_values())

    def test_set_metadata_from_converter_on_resource_info(self):
        converter = MagicMock()
        converter.extract_metadata = MagicMock(return_value={'foo': 'bar'})
        resource_info = ResourceInfo()

        self._create_engine(resource_info=resource_info, converter=converter)
        self.assertEquals({'foo': 'bar'}, resource_info.metadata)

    def test_sets_text_from_converter_on_resource_info(self):
        converter = MagicMock()
        converter.extract_text = MagicMock(return_value=u'foo bar')
        resource_info = ResourceInfo()

        self._create_engine(resource_info=resource_info, converter=converter)
        self.assertEquals(u'foo bar', resource_info.text)

    def test_raises_type_error_for_unknown_extractor_type(self):
        field = Field('foo', extractor=Extractor())
        engine = self._create_engine(fields=[field])

        with self.assertRaises(ExtractionError):
            engine.extract_field_values()

    def test_asserts_proper_type_for_extractors(self):
        field = Field('int_field',
                      extractor=ConstantExtractor('foo'),
                      type_=int)
        engine = self._create_engine(fields=[field])

        with self.assertRaises(ExtractionError):
            engine.extract_field_values()

    def test_asserts_proper_type_for_multivalued_extractors(self):
        field = Field('int_field',
                      extractor=ConstantExtractor([42]),
                      type_=int,
                      multivalued=True)
        engine = self._create_engine(fields=[field])

        self.assertEquals({'int_field': [42]}, engine.extract_field_values())

    def test_provides_default_for_required_fields(self):
        converter = MagicMock()
        converter.extract_metadata = MagicMock(return_value={})

        field = Field('required_field',
                      extractor=ExampleMetadataExtractor(),
                      type_=unicode,
                      required=True)
        engine = self._create_engine(fields=[field], converter=converter)

        self.assertEquals(
            {'required_field': u''}, engine.extract_field_values())

    def test_provides_default_for_required_datetime_fields(self):
        converter = MagicMock()
        converter.extract_metadata = MagicMock(return_value={})

        field = Field('required_datetime',
                      extractor=ExampleMetadataExtractor(),
                      type_=datetime,
                      required=True)
        engine = self._create_engine(fields=[field], converter=converter)

        self.assertEquals(
            {'required_datetime': datetime(1970, 1, 1, 0, 0)},
            engine.extract_field_values())

    def test_skips_field_if_no_value_extracted_and_field_not_required(self):
        converter = MagicMock()
        converter.extract_metadata = MagicMock(return_value={})

        field = Field('optional_field',
                      extractor=ExampleMetadataExtractor(),
                      type_=unicode)
        engine = self._create_engine(fields=[field], converter=converter)

        self.assertEquals({}, engine.extract_field_values())


class TestExtractorBaseClass(CrawlerTestCase):

    def test_extract_value_raises_not_implemented(self):
        extractor = Extractor()
        resource_info = ResourceInfo()
        with self.assertRaises(NotImplementedError):
            extractor.extract_value(resource_info)


class TestPlainTextExtractor(CrawlerTestCase):

    def test_returns_given_text(self):
        extractor = PlainTextExtractor()
        resource_info = ResourceInfo(text=u'foobar')
        extracted_value = extractor.extract_value(resource_info)

        self.assertEquals(u'foobar', extracted_value)
        self.assertIsInstance(extracted_value, unicode)


class TestTitleExtractor(CrawlerTestCase):

    def test_extracts_title_from_x_document_title_http_header(self):
        extractor = TitleExtractor()
        resource_info = ResourceInfo(
            metadata={'title': u'dont-use-this'},
            headers={'X-Document-Title': 'QsOkcmVuZ3JhYmVuCg=='})
        extracted_value = extractor.extract_value(resource_info)

        self.assertEquals(u'B\xe4rengraben', extracted_value)
        self.assertIsInstance(extracted_value, unicode)

    def test_extracts_title_from_metadata(self):
        extractor = TitleExtractor()
        resource_info = ResourceInfo(metadata={'title': u'value'},
                                     headers={})
        extracted_value = extractor.extract_value(resource_info)

        self.assertEquals(u'value', extracted_value)
        self.assertIsInstance(extracted_value, unicode)

    def test_falls_back_to_filename(self):
        extractor = TitleExtractor()
        resource_info = ResourceInfo(
            metadata={},
            headers={'content-disposition': 'attachment; '
                     'filename="document.pdf"'})
        extracted_value = extractor.extract_value(resource_info)

        self.assertEquals(u'document.pdf', extracted_value)
        self.assertIsInstance(extracted_value, unicode)

    def test_falls_back_to_url_slug(self):
        extractor = TitleExtractor()
        resource_info = ResourceInfo(
            metadata={},
            headers={},
            url_info={'loc': 'http://example.org/my____title'})
        extracted_value = extractor.extract_value(resource_info)

        self.assertEquals(u'my-title', extracted_value)
        self.assertIsInstance(extracted_value, unicode)


class TestXpathExtractor(CrawlerTestCase):

    def setUp(self):
        CrawlerTestCase.setUp(self)
        self.extractor = XPathExtractor("//div[@id='content']/h1")

    def _create_resource(self, asset_name):
        doc_fn = resource_filename('ftw.crawler.tests.assets', asset_name)
        resource_info = ResourceInfo(
            metadata={},
            url_info={'loc': 'http//example.org'},
            headers={},
            filename=doc_fn,
            content_type='text/html')
        return resource_info

    def test_extracts_text_from_node_by_xpath_from_html5_doc(self):
        resource_info = self._create_resource('html5_doc.html')
        extracted_value = self.extractor.extract_value(resource_info)

        self.assertEquals(u'Der B\xe4rengraben', extracted_value)
        self.assertIsInstance(extracted_value, unicode)

    def test_extracts_text_from_node_by_xpath_from_xhtml_doc(self):
        resource_info = self._create_resource('xhtml_doc.html')
        extracted_value = self.extractor.extract_value(resource_info)

        self.assertEquals(u'Der B\xe4rengraben', extracted_value)
        self.assertIsInstance(extracted_value, unicode)

    def test_selects_first_if_multiple_nodes_match(self):
        resource_info = self._create_resource('xhtml_doc.html')
        extractor = XPathExtractor("//p")
        extracted_value = extractor.extract_value(resource_info)

        self.assertEquals(u'Foo', extracted_value)
        self.assertIsInstance(extracted_value, unicode)

    def test_raises_if_no_nodes_matched(self):
        resource_info = self._create_resource('xhtml_doc.html')
        extractor = XPathExtractor("//doesntexist")
        with self.assertRaises(NoValueExtracted):
            extractor.extract_value(resource_info)


class TestDescriptionExtractor(CrawlerTestCase):

    def test_extracts_description(self):
        extractor = DescriptionExtractor()
        resource_info = ResourceInfo(metadata={'description': 'value'})
        extracted_value = extractor.extract_value(resource_info)

        self.assertEquals(u'value', extracted_value)
        self.assertIsInstance(extracted_value, unicode)

    def test_raises_if_no_value_found(self):
        extractor = DescriptionExtractor()
        resource_info = ResourceInfo(metadata={})
        with self.assertRaises(NoValueExtracted):
            extractor.extract_value(resource_info)


class TestCreatorExtractor(CrawlerTestCase):

    def test_extracts_creator(self):
        extractor = CreatorExtractor()
        resource_info = ResourceInfo(metadata={'creator': 'John Doe'})
        extracted_value = extractor.extract_value(resource_info)

        self.assertEquals(u'John Doe', extracted_value)
        self.assertIsInstance(extracted_value, unicode)

    def test_raises_if_no_value_found(self):
        extractor = CreatorExtractor()
        resource_info = ResourceInfo(metadata={})
        with self.assertRaises(NoValueExtracted):
            extractor.extract_value(resource_info)


class TestSnippetTextExtractor(CrawlerTestCase):

    def test_returns_plain_text_if_title_not_present(self):
        extractor = SnippetTextExtractor()
        resource_info = ResourceInfo(
            metadata={'title': 'Foo'},
            text='Lorem Ipsum',
            headers={})
        extracted_value = extractor.extract_value(resource_info)

        self.assertEquals(u'Lorem Ipsum', extracted_value)
        self.assertIsInstance(extracted_value, unicode)

    def test_strips_title_from_beginning_of_plain_text(self):
        extractor = SnippetTextExtractor()
        resource_info = ResourceInfo(
            metadata={'title': 'My Title'},
            text='My Title\nLorem Ipsum',
            headers={})
        extracted_value = extractor.extract_value(resource_info)

        self.assertEquals(u'Lorem Ipsum', extracted_value)
        self.assertIsInstance(extracted_value, unicode)

    def test_handles_non_ascii_content(self):
        extractor = SnippetTextExtractor()

        # Both text and title unicode
        resource_info = ResourceInfo(
            metadata={'title': u'B\xe4ren'},
            text=u'B\xe4rengraben',
            headers={})
        extracted_value = extractor.extract_value(resource_info)

        self.assertEquals(u'graben', extracted_value)
        self.assertIsInstance(extracted_value, unicode)

        # Both text and title utf-8
        resource_info = ResourceInfo(
            metadata={'title': 'B\xc3\xa4ren'},
            text='B\xc3\xa4rengraben',
            headers={})
        extracted_value = extractor.extract_value(resource_info)

        self.assertEquals(u'graben', extracted_value)
        self.assertIsInstance(extracted_value, unicode)

        # Mix of unicode and utf-8
        resource_info = ResourceInfo(
            metadata={'title': u'B\xe4ren'},
            text='B\xc3\xa4rengraben',
            headers={})
        extracted_value = extractor.extract_value(resource_info)

        self.assertEquals(u'graben', extracted_value)
        self.assertIsInstance(extracted_value, unicode)


class TestLastModifiedExtractor(DatetimeTestCase):

    def test_lastmod_from_urlinfo(self):
        extractor = LastModifiedExtractor()
        resource_info = ResourceInfo(
            url_info={'lastmod': '2014-12-31T16:45:30+01:00'})

        self.assertEquals(to_utc(datetime(2014, 12, 31, 15, 45, 30)),
                          extractor.extract_value(resource_info))

    def test_falls_back_to_http_last_modified(self):
        extractor = LastModifiedExtractor()
        resource_info = ResourceInfo(
            url_info={},
            headers={'last-modified': 'Wed, 31 Dec 2014 15:45:30 GMT'})
        self.assertEquals(to_utc(datetime(2014, 12, 31, 15, 45, 30)),
                          extractor.extract_value(resource_info))

    def test_falls_back_to_indexing_date(self):
        extractor = LastModifiedExtractor()
        resource_info = ResourceInfo(url_info={}, headers={})
        self.assertDatetimesAlmostEqual(
            datetime.utcnow(), extractor.extract_value(resource_info))


class TestFilenameExtractor(CrawlerTestCase):

    def test_extracts_filename_from_content_disposition(self):
        extractor = FilenameExtractor()
        resource_info = ResourceInfo(
            headers={'content-disposition': 'attachment; '
                     'filename="document.pdf"'})
        extracted_value = extractor.extract_value(resource_info)

        self.assertEquals(u'document.pdf', extracted_value)
        self.assertIsInstance(extracted_value, unicode)

    def test_raises_if_no_content_disposition_header(self):
        extractor = FilenameExtractor()
        resource_info = ResourceInfo(headers={})

        with self.assertRaises(NoValueExtracted):
            extractor.extract_value(resource_info)

    def test_raises_if_header_but_no_filename(self):
        extractor = FilenameExtractor()
        resource_info = ResourceInfo(headers={'content-disposition': ''})

        with self.assertRaises(NoValueExtracted):
            extractor.extract_value(resource_info)


class TestKeywordsExtractor(CrawlerTestCase):

    def test_extracts_comma_separated_keywords(self):
        extractor = KeywordsExtractor()
        resource_info = ResourceInfo(
            metadata={'keywords': 'Foo, Bar,     Baz'})
        extracted_value = extractor.extract_value(resource_info)

        self.assertEquals([u'Foo', u'Bar', u'Baz'], extracted_value)
        for item in extracted_value:
            self.assertIsInstance(item, unicode)

    def test_extracts_whitespace_separated_keywords(self):
        extractor = KeywordsExtractor()
        resource_info = ResourceInfo(metadata={'keywords': u'Foo Bar     Baz'})
        extracted_value = extractor.extract_value(resource_info)

        self.assertEquals([u'Foo', u'Bar', u'Baz'], extracted_value)
        for item in extracted_value:
            self.assertIsInstance(item, unicode)

    def test_raises_if_no_value_found(self):
        extractor = KeywordsExtractor()
        resource_info = ResourceInfo(metadata={})
        with self.assertRaises(NoValueExtracted):
            extractor.extract_value(resource_info)


class TestUIDExtractor(CrawlerTestCase):

    def test_builds_uid_based_on_url(self):
        extractor = UIDExtractor()
        resource_info = ResourceInfo(url_info={'loc': 'http://example.org'})
        extracted_value = extractor.extract_value(resource_info)

        self.assertEquals(u'dab521de-65f9-250b-4cca-7383feef67dc',
                          extracted_value)
        self.assertIsInstance(extracted_value, unicode)

    def test_uid_stays_constant_for_same_url(self):
        extractor = UIDExtractor()
        resource_info = ResourceInfo(url_info={'loc': 'http://example.org'})
        uids = [extractor.extract_value(resource_info) for i in range(10)]
        self.assertEquals(1, len(set(uids)))

    def test_uid_is_different_for_different_urls(self):
        extractor = UIDExtractor()

        resource_info = ResourceInfo(
            url_info={'loc': 'http://example.org'})
        uid1 = extractor.extract_value(resource_info)

        resource_info = ResourceInfo(
            url_info={'loc': 'http://example.org/foo'})
        uid2 = extractor.extract_value(resource_info)

        self.assertNotEqual(uid1, uid2)


class TestSlugExtractor(CrawlerTestCase):

    def test_equals_basename_for_simple_urls(self):
        extractor = SlugExtractor()
        resource_info = ResourceInfo(
            url_info={'loc': 'http://example.org/foo/bar'})
        extracted_value = extractor.extract_value(resource_info)

        self.assertEquals(u'bar', extracted_value)
        self.assertIsInstance(extracted_value, unicode)

    def test_deals_with_trailing_slash(self):
        extractor = SlugExtractor()
        resource_info = ResourceInfo(
            url_info={'loc': 'http://example.org/foo/bar/'})
        extracted_value = extractor.extract_value(resource_info)

        self.assertEquals(u'bar', extracted_value)
        self.assertIsInstance(extracted_value, unicode)

    def test_defaults_to_index_html_for_empty_basename(self):
        extractor = SlugExtractor()
        resource_info = ResourceInfo(
            url_info={'loc': 'http://example.org/'})
        extracted_value = extractor.extract_value(resource_info)

        self.assertEquals(u'index-html', extracted_value)
        self.assertIsInstance(extracted_value, unicode)

    def test_deals_with_url_encoding(self):
        extractor = SlugExtractor()
        resource_info = ResourceInfo(
            url_info={'loc': 'http://example.org/foo%%20bar'})
        extracted_value = extractor.extract_value(resource_info)

        self.assertEquals(u'foo-bar', extracted_value)
        self.assertIsInstance(extracted_value, unicode)

    def test_deals_with_non_ascii_characters_utf8(self):
        extractor = SlugExtractor()
        resource_info = ResourceInfo(
            url_info={'loc': 'http://example.org/b\xc3\xa4rengraben'})
        extracted_value = extractor.extract_value(resource_info)

        self.assertEquals(u'barengraben', extracted_value)
        self.assertIsInstance(extracted_value, unicode)

    def test_deals_with_non_ascii_characters_unicode(self):
        extractor = SlugExtractor()
        resource_info = ResourceInfo(
            url_info={'loc': u'http://example.org/b\xe4rengraben'})
        extracted_value = extractor.extract_value(resource_info)

        self.assertEquals(u'barengraben', extracted_value)
        self.assertIsInstance(extracted_value, unicode)


class TestURLExtractor(CrawlerTestCase):

    def test_extracts_url_from_urlinfo(self):
        extractor = URLExtractor()
        resource_info = ResourceInfo(url_info={'loc': 'http://example.org'})
        extracted_value = extractor.extract_value(resource_info)

        self.assertEquals(u'http://example.org', extracted_value)
        self.assertIsInstance(extracted_value, unicode)


class TestTargetURLExtractor(CrawlerTestCase):

    def test_extracts_target_url_from_urlinfo(self):
        extractor = TargetURLExtractor()
        resource_info = ResourceInfo(url_info={
            'loc': 'http://example.org',
            'target': 'http://example.org/target',
        })
        extracted_value = extractor.extract_value(resource_info)

        self.assertEquals(u'http://example.org/target', extracted_value)
        self.assertIsInstance(extracted_value, unicode)

    def test_defaults_to_loc_if_no_target_given(self):
        extractor = TargetURLExtractor()
        resource_info = ResourceInfo(url_info={'loc': 'http://example.org'})
        extracted_value = extractor.extract_value(resource_info)

        self.assertEquals(u'http://example.org', extracted_value)
        self.assertIsInstance(extracted_value, unicode)


class TestConstantExtractor(CrawlerTestCase):

    def test_returns_constant_value(self):
        extractor = ConstantExtractor(42)
        field = Field('example', extractor)
        extractor.bind(field)
        resource_info = ResourceInfo()

        self.assertEquals(42, extractor.extract_value(resource_info))

    def test_returns_unicode_for_string_constant(self):
        extractor = ConstantExtractor('foo')
        field = Field('example', extractor)
        extractor.bind(field)
        resource_info = ResourceInfo()
        extracted_value = extractor.extract_value(resource_info)

        self.assertEquals(u'foo', extracted_value)
        self.assertIsInstance(extracted_value, unicode)

    def test_returns_unicode_for_multivalued_string_constant(self):
        extractor = ConstantExtractor(['foo', 'bar'])
        field = Field('example', extractor, multivalued=True)
        extractor.bind(field)
        resource_info = ResourceInfo()
        extracted_value = extractor.extract_value(resource_info)

        self.assertEquals([u'foo', u'bar'], extracted_value)
        for item in extracted_value:
            self.assertIsInstance(item, unicode)


class TestIndexingTimeExtractor(DatetimeTestCase):

    def test_returns_current_time(self):
        extractor = IndexingTimeExtractor()
        resource_info = ResourceInfo()
        self.assertDatetimesAlmostEqual(
            datetime.utcnow(),
            extractor.extract_value(resource_info))


class TestSiteAttributeExtractor(CrawlerTestCase):

    def test_retrieves_attribute_from_site(self):
        site = Site('http://example.org', attributes={'name': 'My Site'})
        extractor = SiteAttributeExtractor('name')
        resource_info = ResourceInfo(site=site)

        extracted_value = extractor.extract_value(resource_info)

        self.assertEquals(u'My Site', extracted_value)
        self.assertIsInstance(extracted_value, unicode)

    def test_raises_if_attribute_not_found(self):
        site = Site('http://example.org')
        extractor = SiteAttributeExtractor('name')
        resource_info = ResourceInfo(site=site)

        with self.assertRaises(NoValueExtracted):
            extractor.extract_value(resource_info)


class TestHeaderMappingExtractor(CrawlerTestCase):

    def test_maps_header_to_value(self):
        mapping = {'text/html': 'HTML', 'image/png': 'IMAGE'}
        extractor = HeaderMappingExtractor('content-type', mapping)

        resource_info = ResourceInfo(headers={'content-type': 'text/html'})
        extracted_value = extractor.extract_value(resource_info)
        self.assertEquals('HTML', extracted_value)
        self.assertIsInstance(extracted_value, unicode)

        resource_info = ResourceInfo(headers={'content-type': 'image/png'})
        extracted_value = extractor.extract_value(resource_info)
        self.assertEquals('IMAGE', extracted_value)
        self.assertIsInstance(extracted_value, unicode)

    def test_uses_default_if_header_not_found(self):
        extractor = HeaderMappingExtractor(
            'content-type', {}, default='DEFAULT')
        resource_info = ResourceInfo(headers={})
        extracted_value = extractor.extract_value(resource_info)

        self.assertEquals('DEFAULT', extracted_value)
        self.assertIsInstance(extracted_value, unicode)

    def test_uses_default_if_header_not_mapped(self):
        extractor = HeaderMappingExtractor(
            'pragma', {}, default='DEFAULT')
        resource_info = ResourceInfo(headers={'pragma': 'no-cache'})
        extracted_value = extractor.extract_value(resource_info)

        self.assertEquals('DEFAULT', extracted_value)
        self.assertIsInstance(extracted_value, unicode)

    def test_raises_if_no_default_and_header_not_found(self):
        extractor = HeaderMappingExtractor('content-type', {})
        resource_info = ResourceInfo(headers={})

        with self.assertRaises(NoValueExtracted):
            extractor.extract_value(resource_info)

    def test_raises_if_no_default_and_header_not_mapped(self):
        extractor = HeaderMappingExtractor('content-type', {})
        resource_info = ResourceInfo(headers={'content-type': 'text/html'})

        with self.assertRaises(NoValueExtracted):
            extractor.extract_value(resource_info)

    def test_deals_with_charset_in_content_type_header(self):
        mapping = {'text/html': 'HTML'}
        extractor = HeaderMappingExtractor('content-type', mapping)
        resource_info = ResourceInfo(
            headers={'content-type': 'text/html; charset=utf-8'})
        extracted_value = extractor.extract_value(resource_info)

        self.assertEquals('HTML', extracted_value)
        self.assertIsInstance(extracted_value, unicode)


class TestFieldMappingExtractor(CrawlerTestCase):

    def setUp(self):
        CrawlerTestCase.setUp(self)
        # TODO: Refactor this testcase
        site = Site('http://example.org')
        self.resource_info = ResourceInfo()
        self.mapping = {'travel': 'TRAVEL', 'music': 'MUSIC'}

        subcategory = Field(
            'subcategory',
            extractor=ConstantExtractor('travel'))

        category = Field(
            'category',
            extractor=FieldMappingExtractor('subcategory', self.mapping))

        self.config = Config(
            sites=[site],
            tika=None,
            solr=None,
            unique_field=None,
            url_field=None,
            last_modified_field=None,

            fields=[category, subcategory],)

    def test_maps_field_to_value(self):
        extractor = self.config.get_field('category').extractor
        extracted_value = extractor.extract_value(self.resource_info)

        self.assertEquals(u'TRAVEL', extracted_value)
        self.assertIsInstance(extracted_value, unicode)

    def test_raises_if_field_not_found(self):
        category = self.config.get_field('category')
        category.extractor = FieldMappingExtractor(
            'missing_field', self.mapping)
        category.extractor.bind(category)

        with self.assertRaises(NoSuchField):
            category.extractor.extract_value(self.resource_info)

    def test_uses_default_if_field_returns_none(self):
        category = self.config.get_field('category')
        category.extractor.default = 'DEFAULT'
        subcategory = self.config.get_field('subcategory')
        subcategory.extractor = ConstantExtractor(None)
        subcategory.extractor.bind(subcategory)
        extracted_value = category.extractor.extract_value(self.resource_info)

        self.assertEquals(u'DEFAULT', extracted_value)
        self.assertIsInstance(extracted_value, unicode)

    def test_raises_if_no_default_and_field_doesnt_return_value(self):
        category = self.config.get_field('category')
        subcategory = self.config.get_field('subcategory')
        subcategory.extractor = ConstantExtractor(None)
        subcategory.extractor.bind(subcategory)

        with self.assertRaises(NoValueExtracted):
            category.extractor.extract_value(self.resource_info)

    def test_uses_default_if_field_value_not_mapped(self):
        category = self.config.get_field('category')
        category.extractor.default = 'DEFAULT'
        subcategory = self.config.get_field('subcategory')
        subcategory.extractor = ConstantExtractor('physics')
        subcategory.extractor.bind(subcategory)
        extracted_value = category.extractor.extract_value(self.resource_info)

        self.assertEquals(u'DEFAULT', extracted_value)
        self.assertIsInstance(extracted_value, unicode)

    def test_raises_if_no_default_and_field_value_not_mapped(self):
        category = self.config.get_field('category')
        subcategory = self.config.get_field('subcategory')
        subcategory.extractor = ConstantExtractor('physics')
        subcategory.extractor.bind(subcategory)

        with self.assertRaises(NoValueExtracted):
            category.extractor.extract_value(self.resource_info)



from ftw.crawler.testing import XMLTestCase
from ftw.crawler.tests.helpers import get_asset
from ftw.crawler.xml_utils import remove_namespaces
from lxml import etree
import io


NAMESPACED_XML = get_asset('namespaced.xml')


class TestRemoveNamespaces(XMLTestCase):

    def test_removes_namespaces(self):
        namespaced_tree = etree.parse(io.BytesIO(NAMESPACED_XML))
        tree = remove_namespaces(namespaced_tree)
        output_xml = etree.tostring(
            tree, xml_declaration=True, encoding='utf-8')
        expected = """\
        <?xml version='1.0' encoding='utf-8'?>
        <root>
            <node>
                <node>Text</node>
            </node>
        </root>
        """
        self.assertXMLEquals(output_xml, expected)



from ftw.crawler.utils import mkdir_p
from logging.handlers import TimedRotatingFileHandler
from logging import Formatter
from logging import StreamHandler
import logging
import os
import sys


LOGDIR = 'var/log/'
CONSOLE_FORMAT = '%(levelname)-8s %(module)-10s %(message)s'
FILE_FORMAT = '%(asctime)s %(levelname)-8s %(module)-10s %(message)s'
file_formatter = Formatter(FILE_FORMAT, datefmt='%Y-%m-%d %H:%M:%S')


def create_log_dir():
    script_path = os.path.abspath(sys.argv[0])
    bin_dir = os.path.dirname(script_path)
    if bin_dir.endswith('bin'):
        # Use the /bin directory's parent as base
        base_dir = os.path.split(bin_dir)[0]
    else:
        # Unless there is no /bin - default to cwd
        base_dir = os.getcwd()

    # Create log directory relative to base
    log_dir = os.path.join(base_dir, LOGDIR)
    mkdir_p(log_dir)
    return log_dir


def setup_logging():
    log_dir = create_log_dir()
    logging.root.setLevel(logging.DEBUG)

    # StreamHandler logging to console (stderr) at level INFO
    console_handler = StreamHandler()
    console_handler.setLevel(logging.INFO)
    console_handler.setFormatter(Formatter(CONSOLE_FORMAT))
    logging.root.addHandler(console_handler)

    # FileHandler logging to 'debug.log' at level DEBUG
    debug_file_handler = TimedRotatingFileHandler(
        os.path.join(log_dir, 'debug.log'), when='midnight', backupCount=30)
    debug_file_handler.setLevel(logging.DEBUG)
    debug_file_handler.setFormatter(file_formatter)
    logging.root.addHandler(debug_file_handler)

    # FileHandler logging to 'info.log' at level INFO
    info_file_handler = TimedRotatingFileHandler(
        os.path.join(log_dir, 'info.log'), when='midnight', backupCount=30)
    info_file_handler.setLevel(logging.INFO)
    info_file_handler.setFormatter(file_formatter)
    logging.root.addHandler(info_file_handler)

    # FileHandler logging to 'warn.log' at level WARN
    warn_file_handler = TimedRotatingFileHandler(
        os.path.join(log_dir, 'warn.log'), when='midnight', backupCount=30)
    warn_file_handler.setLevel(logging.WARN)
    warn_file_handler.setFormatter(file_formatter)
    logging.root.addHandler(warn_file_handler)

    # Only log messages from the requests library with at least level WARN
    requests_logger = logging.getLogger('requests')
    requests_logger.setLevel(logging.WARN)



from ftw.crawler.metadata import SimpleMetadata
from ftw.crawler.testing import CrawlerTestCase


class TestSimpleMetadata(CrawlerTestCase):

    def test_original_prefixed_entries_are_preserved(self):
        metadata = {'dcterms:title': 'My Title'}
        simple_metadata = SimpleMetadata(metadata)
        self.assertEquals('My Title', simple_metadata['dcterms:title'])

    def test_handles_dcterms_title(self):
        metadata = {'dcterms:title': 'My Title'}
        simple_metadata = SimpleMetadata(metadata)
        self.assertEquals('My Title', simple_metadata['title'])

    def test_handles_dc_title(self):
        metadata = {'dc:title': 'My Title'}
        simple_metadata = SimpleMetadata(metadata)
        self.assertEquals('My Title', simple_metadata['title'])

    def test_handles_dcterms_take_precedence_over_dc(self):
        metadata = {'dc:title': 'Some irrelevant Title',
                    'dcterms:title': 'My Title'}
        simple_metadata = SimpleMetadata(metadata)
        self.assertEquals('My Title', simple_metadata['title'])



from ftw.crawler.exceptions import SolrError
from ftw.crawler.solr import solr_escape
from ftw.crawler.solr import SolrConnector
from ftw.crawler.testing import CrawlerTestCase
from ftw.crawler.testing import SolrTestCase
from ftw.crawler.tests.helpers import MockResponse
from mock import patch
import json


class TestSolrConnector(SolrTestCase):

    def setUp(self):
        SolrTestCase.setUp(self)
        self.response_ok = self.create_solr_response()

        docs = [{'Title': 'Foobar'}, {'Title': 'Foo bar'}]
        self.response_search_results = self.create_solr_results(docs)

        self.response_400_bad_request = self.create_solr_response(
            status=400,
            content='{"responseHeader":{"status":400,"QTime":0},'
                    '"error":''{"msg":"Something went wrong","code":400}}')

    @patch('requests.post')
    def test_index_returns_response(self, request):
        request.return_value = self.response_ok

        solr = SolrConnector('http://localhost:8983/solr')
        response = solr.index({'field': 'value'})
        self.assertIsInstance(response, MockResponse)

    @patch('requests.post')
    def test_delete_returns_response(self, request):
        request.return_value = self.response_ok

        solr = SolrConnector('http://localhost:8983/solr')
        response = solr.delete('12345')
        self.assertIsInstance(response, MockResponse)

    @patch('requests.get')
    def test_search_returns_documents(self, request):
        request.return_value = self.response_search_results
        solr = SolrConnector('http://localhost:8983/solr')

        docs = solr.search('Title:Foo*')
        self.assertEquals([{'Title': 'Foobar'}, {'Title': 'Foo bar'}], docs)

    @patch('requests.post')
    def test_index_sends_proper_request_to_solr(self, request):
        request.return_value = self.response_ok
        solr = SolrConnector('http://localhost:8983/solr')

        expected_url = 'http://localhost:8983/solr/update?commit=true'
        expected_headers = {'Content-Type': 'application/json'}
        data = {'field': 'value'}

        solr.index(data)
        request.assert_called_with(
            expected_url, headers=expected_headers, data=json.dumps([data]))

    @patch('requests.post')
    def test_delete_sends_proper_request_to_solr(self, request):
        request.return_value = self.response_ok
        solr = SolrConnector('http://localhost:8983/solr')

        expected_url = 'http://localhost:8983/solr/update?commit=true'
        expected_headers = {'Content-Type': 'application/json'}
        uid = '12345'
        cmd = {'delete': {'id': uid}}

        solr.delete(uid)
        request.assert_called_with(
            expected_url, headers=expected_headers, data=json.dumps(cmd))

    @patch('requests.get')
    def test_search_sends_proper_request_to_solr(self, request):
        request.return_value = self.response_search_results
        solr = SolrConnector('http://localhost:8983/solr')

        query = 'Title:Foo*'
        expected_url = 'http://localhost:8983/solr/select'
        expected_headers = {'Content-Type': 'application/json'}
        params = {'q': query, 'wt': 'json'}

        solr.search(query)

        request.assert_called_with(
            expected_url, headers=expected_headers, params=params)

    @patch('ftw.crawler.solr.log')
    @patch('requests.post')
    def test_index_logs_non_200_responses_from_solr(self, request, log):
        request.return_value = self.response_400_bad_request

        solr = SolrConnector('http://localhost:8983/solr')
        solr.index({'field': 'value'})
        self.assertTrue(log.error.called)

    @patch('ftw.crawler.solr.log')
    @patch('requests.post')
    def test_delete_logs_non_200_responses_from_solr(self, request, log):
        request.return_value = self.response_400_bad_request

        solr = SolrConnector('http://localhost:8983/solr')
        solr.delete('12345')
        self.assertTrue(log.error.called)

    @patch('ftw.crawler.solr.log')
    @patch('requests.get')
    def test_search_raises_on_non_200_responses_from_solr(self, request, log):
        request.return_value = self.response_400_bad_request
        solr = SolrConnector('http://localhost:8983/solr')

        with self.assertRaises(SolrError):
            solr.search('Title:Foo')
            self.assertTrue(log.error.called)

    @patch('requests.get')
    def test_search_honors_fl_argument(self, request):
        request.return_value = self.response_search_results
        solr = SolrConnector('http://localhost:8983/solr')

        solr.search('Title:Foo', fl=('Title', 'UID'))
        args, kwargs = request.call_args
        self.assertIn(('fl', 'Title,UID'), kwargs['params'].items())


class TestSolrEscape(CrawlerTestCase):

    def test_escapes_special_characters(self):
        value = r'+ - && || ! ( ) { } [ ] ^ " ~ * ? : \ /'
        expected = r'\+ \- \&& \|| \! \( \) \{ \} \[ \] '\
                   r'\^ \" \~ \* \? \: \\ \/'

        self.assertEquals(expected, solr_escape(value))



from argparse import Namespace
from copy import deepcopy
from ftw.crawler.configuration import get_config
from ftw.crawler.purging import purge_removed_docs_from_index
from ftw.crawler.sitemap import VirtualSitemapIndex
from ftw.crawler.testing import SitemapTestCase
from ftw.crawler.testing import SolrTestCase
from mock import patch
from pkg_resources import resource_filename


BASIC_CONFIG = resource_filename('ftw.crawler.tests.assets', 'basic_config.py')


class TestPurging(SolrTestCase, SitemapTestCase):

    def setUp(self):
        SolrTestCase.setUp(self)
        SitemapTestCase.setUp(self)
        args = Namespace(tika=None, solr=None,
                         slacktoken=None, slackchannel=None)
        args.config = BASIC_CONFIG
        self.config = deepcopy(get_config(args))
        self.config.url_field = 'url'

    @patch('ftw.crawler.solr.SolrConnector.delete')
    def test_purges_doc_removed_from_sitemap(self, delete):
        indexed_docs = [
            {'UID': '1', 'url': 'http://www.pctipp.ch/download'},
            {'UID': '2', 'url': 'http://www.pctipp.ch/about'},
        ]
        pctipp = self.config.get_site('http://www.pctipp.ch/')

        sitemap = self.create_sitemap(
            urls=['http://www.pctipp.ch/download'],
            site=pctipp)

        sitemap_index = VirtualSitemapIndex(pctipp, sitemaps=[sitemap])
        purge_removed_docs_from_index(self.config, sitemap_index, indexed_docs)

        delete.assert_called_with(u'2')
        self.assertEqual(1, delete.call_count)

    @patch('ftw.crawler.solr.SolrConnector.delete')
    def test_doesnt_touch_any_docs_not_starting_with_site_urls(self, delete):
        indexed_docs = [
            {'UID': '1', 'url': 'http://www.pctipp.ch/download'},
            {'UID': '2', 'url': 'http://www.pctipp.ch/about'},
            {'UID': '3', 'url': 'http://some.com/other/url'},
        ]
        pctipp = self.config.get_site('http://www.pctipp.ch/')

        sitemap = self.create_sitemap(
            urls=['http://www.pctipp.ch/download',
                  'http://www.pctipp.ch/about'],
            site=pctipp)

        sitemap_index = VirtualSitemapIndex(pctipp, sitemaps=[sitemap])
        purge_removed_docs_from_index(self.config, sitemap_index, indexed_docs)

        self.assertEquals(0, delete.call_count)



from ftw.crawler.exceptions import SolrError
from ftw.crawler.utils import ExtendedJSONEncoder
import logging
import requests


log = logging.getLogger(__name__)


# Solr/lucene reserved characters/terms:
# + - && || ! ( ) { } [ ] ^ " ~ * ? : \ /
SPECIAL_TOKENS = ['+', '-', '&&', '||', '!', '(', ')', '{', '}', '[', ']',
                  '^', '"', '~', '*', '?', ':', '\\', '/']


def solr_escape(value):
    """Escape a value for use in a Solr/Lucene query
    """
    # Deal with backslashes first
    value = value.replace('\\', '\\\\')
    for token in SPECIAL_TOKENS:
        if not token == '\\':
            value = value.replace(token, '\\' + token)
    return value


class SolrConnector(object):

    update_handler = 'update'
    search_handler = 'select'

    def __init__(self, solr_base):
        self.solr_base = solr_base.rstrip('/')

        self.update_url = '{}/{}?{}'.format(
            self.solr_base, SolrConnector.update_handler, 'commit=true')

        self.search_url = '{}/{}'.format(
            self.solr_base, SolrConnector.search_handler)

    def _update_request(self, data):
        headers = {'Content-Type': 'application/json'}
        document = ExtendedJSONEncoder().encode(data)
        response = requests.post(
            self.update_url, data=document, headers=headers)

        if not response.status_code == 200:
            log.error(u"Error from Solr: Status: {}. Response:\n{}".format(
                response.status_code, response.text))
        return response

    def _search_request(self, query, fl=None):
        headers = {'Content-Type': 'application/json'}

        params = {'q': query, 'wt': 'json'}

        if fl is not None:
            params.update({'fl': ','.join(fl)})

        response = requests.get(
            self.search_url, params=params, headers=headers)

        if not response.status_code == 200:
            log.error(u"Error from Solr: Status: {}. Response:\n{}".format(
                response.status_code, response.text))
            raise SolrError(
                "Got status {} from Solr.".format(response.status_code))
        return response

    def index(self, document):
        response = self._update_request([document])
        return response

    def delete(self, unique_id):
        del_command = {'delete': {'id': unique_id}}
        response = self._update_request(del_command)
        return response

    def search(self, query, fl=None):
        response = self._search_request(query, fl)
        search_results = response.json()['response']
        docs = search_results['docs']
        return docs



from setuptools import setup, find_packages
import os

version = '1.2.0'

tests_require = [
    'unittest2',
    'mock',
    'testfixtures',
    'ftw.crawler [slack]',
]

slack_requires = [
    'slacker',
]

setup(name='ftw.crawler',
      version=version,
      description='Crawl sites, extract text and metadata, index it in Solr',
      long_description=open('README.rst').read() + '\n' +
      open(os.path.join('docs', 'HISTORY.txt')).read(),

      classifiers=[
          'Intended Audience :: Developers',
          'License :: OSI Approved :: GNU General Public License (GPL)',
          'Programming Language :: Python',
          'Topic :: Software Development',
      ],

      keywords='crawling extraction solr',
      author='4teamwork AG',
      author_email='mailto:info@4teamwork.ch',
      url='https://github.com/4teamwork/ftw.crawler',
      license='GPL2',

      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['ftw'],
      include_package_data=True,
      zip_safe=False,

      install_requires=[
          'setuptools',
          'requests[security]',
          'lxml',
          'python-dateutil',
          'pytz',
          'python-slugify',
          'BeautifulSoup',
          'chardet'
      ],

      tests_require=tests_require,
      extras_require=dict(tests=tests_require, slack=slack_requires),

      entry_points='''
      [console_scripts]
      crawl = ftw.crawler.main:main
      ''',
      )



from ftw.crawler.exceptions import NoSuchField
from ftw.crawler.exceptions import SiteNotFound
import imp
import os


def get_config(options):
    config_path = os.path.abspath(options.config)
    module_name = os.path.splitext(os.path.basename(options.config))[0]
    module = imp.load_source(module_name, config_path)
    config = module.CONFIG

    # Command line overrides for config options
    if options.tika:
        config.tika = options.tika
    if options.solr:
        config.solr = options.solr
    if options.slacktoken:
        config.slacktoken = options.slacktoken
    if options.slackchannel:
        config.slackchannel = options.slackchannel

    if not (config.tika and config.solr):
        raise ValueError(
            'Tika and Solr URLs must be specified, either via command line '
            'arguments or parameters in the configuration file.')

    return module.CONFIG


class Config(object):

    def __init__(self, sites, unique_field, url_field, last_modified_field,
                 fields, tika=None, solr=None,
                 slacktoken=None, slackchannel=None):
        self.sites = sites
        self.unique_field = unique_field
        self.url_field = url_field
        self.last_modified_field = last_modified_field
        self.fields = fields
        self.tika = tika
        self.solr = solr
        self.slacktoken = slacktoken
        self.slackchannel = slackchannel

        for site in self.sites:
            site.bind(self)

        for field in self.fields:
            field.bind(self)

    def get_field(self, field_name):
        for field in self.fields:
            if field.name == field_name:
                return field
        raise NoSuchField(field_name)

    def get_site(self, url):
        for site in self.sites:
            if site.url == url:
                return site
        raise SiteNotFound("Couldn't find site %r in config!" % url)


class Site(object):

    def __init__(self, url, attributes=None, sleeptime=0.1):
        self.url = url
        self.sleeptime = sleeptime

        if attributes is None:
            attributes = {}
        self.attributes = attributes

    def bind(self, config):
        self.config = config


class Field(object):

    def __init__(self, name, extractor, type_=unicode, required=False,
                 multivalued=False):
        self.name = name
        self.extractor = extractor
        self.type_ = type_
        self.required = required
        self.multivalued = multivalued

        self.extractor.bind(self)

    def bind(self, config):
        self.config = config

    def __repr__(self):
        desc = "<Field '{}' type_={} required={} multivalued={} extractor={}>"
        return desc.format(
            self.name, self.type_.__name__, self.required, self.multivalued,
            self.extractor)



from urlparse import urlsplit
from wsgiref.handlers import format_date_time
import calendar
import datetime
import dateutil.parser
import errno
import gzip
import io
import json
import os
import pytz


def to_utc(dt):
    """Ensure that a datetime object is time zone aware and in UTC.
    """
    if dt.tzinfo is None:
        # Naive datetime, assume UTC
        utc_date = pytz.utc.localize(dt)
    else:
        # TZ aware datetime. Convert to UTC if necessary
        utc_date = dt.astimezone(pytz.utc)
    return utc_date


def to_iso_datetime(dt):
    """Create a valid, time zone aware ISO 8601 date/time string in UTC.
    """
    fmt = "%Y-%m-%dT%H:%M:%S.%fZ"
    return to_utc(dt).strftime(fmt)


def from_iso_datetime(datestring):
    """Parse an ISO 8601 datetime and create a TZ aware datetime object in UTC.
    """
    dt = dateutil.parser.parse(datestring)
    return to_utc(dt)


def to_http_datetime(dt):
    """Create a valid, time zone aware RFC 2616 HTTP datetime string in GMT.
    """
    dt = to_utc(dt)
    # Use calendar.timegm(), NOT time.mktime(), which assumes *local time*
    timestamp = calendar.timegm(dt.timetuple())
    return format_date_time(timestamp)


def from_http_datetime(datestring):
    """Parse an RFC 2616 HTTP datetime string and create a TZ aware datetime
    object in UTC.
    """
    dt = dateutil.parser.parse(datestring)
    return to_utc(dt)


def get_content_type(header_value):
    """Helps deal with the fact that the HTTP Content-Type header may also
    contain a charset declaration. For example:

    text/html; charset=utf-8

    More often than not, we're only interested in the MIME type though.
    """
    if header_value is not None:
        return header_value.split(';')[0]


def is_gzipped(response):
    """Determine whether a response's content is gzipped.

    This only considers the Content-Type header and the filename, NOT
    HTTP compression indicated by the Content-Encoding header, which is
    handled transparently by the `requests` module.
    """
    content_type = get_content_type(response.headers.get('Content-Type'))
    path = urlsplit(response.request.url).path
    return content_type == 'application/x-gzip' or path.endswith('.gz')


def gunzip(bytestring):
    """Decompress a gzipped bytestring.
    """
    with gzip.GzipFile(mode='rb', fileobj=io.BytesIO(bytestring)) as f:
        return f.read()


class ExtendedJSONEncoder(json.JSONEncoder):
    """JSONEncoder that can also serialize datetime objects.
    """
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return to_iso_datetime(obj)
        else:
            return super(ExtendedJSONEncoder, self).default(obj)


def normalize_whitespace(s):
    """Normalize whitespace in a string by
    - replacing all occurences of CR, LF and TAB with a space
    - replacing multiple spaces with only one
    - removing any leading or trailing whitespace
    """
    s = safe_unicode(s)
    return u' '.join(s.split())


def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def safe_unicode(value):
    if isinstance(value, str):
        return value.decode('utf-8')
    return value



from datetime import datetime
from ftw.crawler.configuration import Site
from ftw.crawler.exceptions import AttemptedRedirect
from ftw.crawler.exceptions import FetchingError
from ftw.crawler.exceptions import NotModified
from ftw.crawler.resource import ResourceInfo
from ftw.crawler.testing import FetcherTestCase
from ftw.crawler.tests.helpers import MockResponse
from ftw.crawler.utils import to_http_datetime
from ftw.crawler.utils import to_iso_datetime
from ftw.crawler.utils import to_utc
from mock import patch
from testfixtures import LogCapture
import logging
import shutil
import tempfile


class TestResourceFetcher(FetcherTestCase):

    def setUp(self):
        FetcherTestCase.setUp(self)
        self.tempdir = tempfile.mkdtemp(prefix='ftw.crawler.tests_')

    def tearDown(self):
        FetcherTestCase.tearDown(self)
        shutil.rmtree(self.tempdir)

    @patch('requests.sessions.Session.get')
    def test_fetches_and_saves_resource(self, request):
        request.return_value = MockResponse(
            content='MARKER', headers={'Content-Type': 'text/html'})
        resource_info = ResourceInfo(url_info={'loc': 'http://example.org/'})

        fetcher = self._create_fetcher(resource_info, tempdir=self.tempdir)
        resource_info = fetcher.fetch()

        self.assertTrue(resource_info.filename.startswith(self.tempdir))

        self.assertEquals('text/html', resource_info.content_type)
        with open(resource_info.filename) as resource_file:
            self.assertEquals('MARKER', resource_file.read())

    @patch('requests.sessions.Session.get')
    def test_returns_http_headers(self, request):
        request.return_value = MockResponse(
            content='', headers={'Content-Type': 'text/html'})
        resource_info = ResourceInfo(url_info={'loc': 'http://example.org/'})
        fetcher = self._create_fetcher(resource_info=resource_info)
        resource_info = fetcher.fetch()

        self.assertEquals({'Content-Type': 'text/html'}, resource_info.headers)

    @patch('requests.sessions.Session.get')
    def test_raises_if_not_200_ok(self, request):
        request.return_value = MockResponse(status_code=404)

        resource_info = ResourceInfo(url_info={'loc': 'http://example.org/'})
        fetcher = self._create_fetcher(resource_info)
        with self.assertRaises(FetchingError):
            fetcher.fetch()

    @patch('requests.sessions.Session.get')
    def test_sleeps_and_retries_when_too_many_requests(self, request):
        responses = [
            MockResponse(status_code=429),
            MockResponse(status_code=429),
            MockResponse(content='', headers={'Content-Type': 'text/html'}),
        ]
        request.side_effect = lambda url, **kw: responses.pop(0)

        resource_info = ResourceInfo(url_info={'loc': 'http://example.org/'},
                                     site=Site('http://example.org/'))
        fetcher = self._create_fetcher(resource_info)
        logging.disable(logging.INFO)
        with LogCapture() as log:
            resource_info = fetcher.fetch()

        log.check(('ftw.crawler.fetcher',
                   'WARNING',
                   u'429 Too Many Requests, sleeping for 0.1s'),
                  ('ftw.crawler.fetcher',
                   'WARNING',
                   u'429 Too Many Requests, sleeping for 0.2s'))

        self.assertEquals({'Content-Type': 'text/html'},
                          resource_info.headers)

    @patch('requests.sessions.Session.get')
    def test_raises_if_redirect(self, request):
        request.return_value = MockResponse(status_code=301, is_redirect=True)

        resource_info = ResourceInfo(url_info={'loc': 'http://example.org/'})
        fetcher = self._create_fetcher(resource_info)
        with self.assertRaises(AttemptedRedirect):
            fetcher.fetch()

    @patch('requests.sessions.Session.get')
    def test_doesnt_choke_on_charset_in_content_type(self, request):
        request.return_value = MockResponse(
            content='', headers={'Content-Type': 'text/html; charset=utf-8'})

        resource_info = ResourceInfo(url_info={'loc': 'http://example.org/'})
        fetcher = self._create_fetcher(resource_info)
        resource_info = fetcher.fetch()

        self.assertEquals('text/html', resource_info.content_type)


class TestFetcherIsModifiedLogic(FetcherTestCase):

    def _create_resource_info(self):
        return ResourceInfo(
            url_info={'loc': 'http://example.org/'},
            last_indexed=None)

    def test_unknown_last_indexed_time_always_yields_modified(self):
        fetcher = self._create_fetcher(self._create_resource_info())
        self.assertTrue(
            fetcher.is_modified(),
            "Unknown last_indexed time should result in resource being "
            "considered MODIFIED")

    def test_is_modified_tests_against_urlinfo_lastmod(self):
        resource_info = self._create_resource_info()
        fetcher = self._create_fetcher(resource_info)
        resource_info.last_indexed = to_utc(datetime(2014, 1, 1, 15, 30))

        server_modified = to_utc(datetime(2014, 1, 1, 15, 30))
        resource_info.url_info['lastmod'] = to_iso_datetime(server_modified)

        self.assertFalse(
            fetcher.is_modified(),
            "Equal modification dates should lead to resource being "
            "considered UNMODIFIED")

        server_modified = to_utc(datetime(2014, 1, 1, 15, 31))
        resource_info.url_info['lastmod'] = to_iso_datetime(server_modified)

        self.assertTrue(
            fetcher.is_modified(),
            "Newer server modification date should lead to resource being "
            "considered MODIFIED")

        server_modified = to_utc(datetime(2014, 1, 1, 15, 29))
        resource_info.url_info['lastmod'] = to_iso_datetime(server_modified)

        self.assertFalse(
            fetcher.is_modified(),
            "Older server modification date should lead to resource being "
            "considered UNMODIFIED")

    @patch('requests.sessions.Session.head')
    def test_is_modified_tests_against_http_last_modified(self, request):
        resource_info = self._create_resource_info()
        fetcher = self._create_fetcher(resource_info)
        resource_info.last_indexed = to_utc(datetime(2014, 1, 1, 15, 30))

        server_modified = to_utc(datetime(2014, 1, 1, 15, 30))
        request.return_value = MockResponse(
            headers={'last-modified': to_http_datetime(server_modified)})

        self.assertFalse(
            fetcher.is_modified(),
            "Equal modification dates should lead to resource being "
            "considered UNMODIFIED")

        server_modified = to_utc(datetime(2014, 1, 1, 15, 31))
        request.return_value = MockResponse(
            headers={'last-modified': to_http_datetime(server_modified)})

        self.assertTrue(
            fetcher.is_modified(),
            "Newer server modification date should lead to resource being "
            "considered MODIFIED")

        server_modified = to_utc(datetime(2014, 1, 1, 15, 29))
        request.return_value = MockResponse(
            headers={'last-modified': to_http_datetime(server_modified)})

        self.assertFalse(
            fetcher.is_modified(),
            "Older server modification date should lead to resource being "
            "considered UNMODIFIED")

    @patch('requests.sessions.Session.head')
    def test_is_modified_defaults_to_true(self, request):
        resource_info = self._create_resource_info()
        fetcher = self._create_fetcher(resource_info)
        resource_info.last_indexed = to_utc(datetime(2014, 1, 1, 15, 30))

        self.assertTrue(
            fetcher.is_modified(),
            "is_modified() should default to True if last_indexed date "
            "is present, but no server modification date could be determined")

    @patch('ftw.crawler.fetcher.ResourceFetcher.is_modified')
    @patch('requests.sessions.Session.get')
    def test_fetcher_doesnt_fetch_if_not_modified(self, request, is_modified):
        is_modified.return_value = False
        resource_info = self._create_resource_info()
        fetcher = self._create_fetcher(resource_info)
        with self.assertRaises(NotModified):
            fetcher.fetch()
        self.assertEquals(0, request.call_count)



from BeautifulSoup import UnicodeDammit
from datetime import datetime
from ftw.crawler.exceptions import ExtractionError
from ftw.crawler.exceptions import NoValueExtracted
from ftw.crawler.utils import from_iso_datetime
from ftw.crawler.utils import get_content_type
from ftw.crawler.utils import normalize_whitespace
from ftw.crawler.utils import safe_unicode
from ftw.crawler.xml_utils import MARKUP_TYPES
from ftw.crawler.xml_utils import remove_namespaces
from lxml import etree
from slugify import slugify
from urllib import unquote_plus
from urlparse import urlparse
from uuid import UUID
import hashlib
import logging
import lxml.html


log = logging.getLogger(__name__)


class Extractor(object):
    """Base class for all extractors.
    """

    def extract_value(self, resource_info):
        raise NotImplementedError

    def bind(self, field):
        self.field = field

    def __repr__(self):
        cls = self.__class__
        name = '.'.join((cls.__module__, cls.__name__))
        return "<{}>".format(name)


class MetadataExtractor(Extractor):
    """Base class for all extractors that extract structured metadata via Tika.
    """


class TextExtractor(Extractor):
    """Base class for all extractors that extract plain text via Tika.
    """


class TextFromMarkupExtractor(Extractor):
    """Base class for all extractors that extract text from markup content.
    """


class URLInfoExtractor(Extractor):
    """Base class for all extractors that extract data from the sitemap's
    <urlinfo /> structure.
    """


class HTTPHeaderExtractor(Extractor):
    """Base class for all extractors that extract data from the response's
    HTTP headers.
    """


class ResourceIndependentExtractor(Extractor):
    """Base class for all extractors that don't need any information from
    the resource whatsoever.
    """


class SiteConfigExtractor(Extractor):
    """Base class for all extractors that extract data based on the `Site`
    configuration object.
    """


class ExtractionEngine(object):

    extractor_types = (
        MetadataExtractor, TextExtractor, URLInfoExtractor,
        ResourceIndependentExtractor, SiteConfigExtractor, HTTPHeaderExtractor,
        TextFromMarkupExtractor
    )

    def __init__(self, config, resource_info, converter):
        self.config = config
        self.resource_info = resource_info

        self.resource_info.metadata = converter.extract_metadata(
            self.resource_info)

        self.resource_info.text = safe_unicode(converter.extract_text(
            self.resource_info))

    def _unkown_extractor_type(self, extractor):
        cls = extractor.__class__
        raise ExtractionError(
            "Unknown extractor type for '{}' - must inherit from at least one "
            "of {}. (Current base classes: {})".format(
                ExtractionEngine.extractor_types, cls, cls.__bases__))

    def _assert_proper_type(self, field, value, extractor):
        if field.multivalued:
            valid_type = all(isinstance(v, field.type_) for v in value)
        else:
            valid_type = isinstance(value, field.type_)
        if not valid_type:
            raise ExtractionError(
                "Invalid return value type '{}' for extractor {} and field "
                "{}. Return value was: {}".format(
                    type(value).__name__, extractor, field, repr(value)))

    def _get_field_default(self, field):
        type_ = field.type_
        if issubclass(type_, datetime):
            epoch = datetime.utcfromtimestamp(0)
            return epoch
        else:
            # Return zero value for the respective type by instantiating it
            return type_()

    def extract_field_values(self):
        field_values = {}
        for field in self.config.fields:
            extractor = field.extractor

            if not isinstance(extractor, ExtractionEngine.extractor_types):
                self._unkown_extractor_type(extractor)

            try:
                value = extractor.extract_value(self.resource_info)
            except NoValueExtracted:
                if field.required:
                    value = self._get_field_default(field)
                else:
                    # No value could be extracted, and field is not required,
                    # so we may skip this field
                    continue
            self._assert_proper_type(field, value, extractor)
            field_values.update({field.name: value})
        return field_values


class PlainTextExtractor(TextExtractor):

    def extract_value(self, resource_info):
        return normalize_whitespace(resource_info.text)


class UIDExtractor(URLInfoExtractor):

    def extract_value(self, resource_info):
        url = resource_info.url_info['loc']
        hash_ = hashlib.md5(url)
        uid = UUID(bytes=hash_.digest())
        return unicode(uid)


class SlugExtractor(URLInfoExtractor):

    def _make_slug(self, value):
        value = unquote_plus(value)
        if not isinstance(value, unicode):
            value = value.decode('utf-8')
        slug = slugify(value)
        return slug

    def extract_value(self, resource_info):
        url = resource_info.url_info.get('loc')
        path = urlparse(url).path.rstrip('/')
        basename = path.split('/')[-1]
        if basename == '':
            basename = 'index-html'
        slug = self._make_slug(basename)
        return safe_unicode(slug)


class URLExtractor(URLInfoExtractor):

    def extract_value(self, resource_info):
        url = resource_info.url_info.get('loc')
        return safe_unicode(url)


class TargetURLExtractor(URLInfoExtractor):

    def extract_value(self, resource_info):
        if 'target' in resource_info.url_info:
            return safe_unicode(resource_info.url_info['target'])
        else:
            return URLExtractor().extract_value(resource_info)


class TitleExtractor(MetadataExtractor, HTTPHeaderExtractor, URLInfoExtractor):

    def _extract_title(self, resource_info):
        # If present, X-Document-Title header takes precedence
        if 'X-Document-Title' in resource_info.headers:
            header_value = resource_info.headers['X-Document-Title']
            return header_value.decode('base64').decode('utf-8').strip()

        # Next, try to get title from a `div#content h1` element
        h1_extractor = XPathExtractor("//div[@id='content']/h1")
        try:
            value = h1_extractor.extract_value(resource_info)
            return value
        except NoValueExtracted:
            pass

        # Next, attempt to get a title from Tika metadata
        value = resource_info.metadata.get('title')

        if value is None:
            try:
                # Fall back to filename from Content-Disposition header
                value = FilenameExtractor().extract_value(resource_info)
            except NoValueExtracted:
                # As a last resort, use a slug built from the rightmost part
                # of the resource's URL
                value = SlugExtractor().extract_value(resource_info)

        return value

    def extract_value(self, resource_info):
        title = self._extract_title(resource_info)
        return normalize_whitespace(title)


class XPathExtractor(TextFromMarkupExtractor, URLInfoExtractor):

    def __init__(self, xpath):
        self.xpath = xpath

    def _sniff_encoding(self, resource_info):
        with open(resource_info.filename) as f:
            data = f.read()
            proposed = ["utf-8", "latin1"]
            converted = UnicodeDammit(data, proposed, isHTML=True)
        del data
        return converted.originalEncoding

    def _get_tree(self, resource_info, encoding):
        # Use HTMLParser for everything, even XML / XHTML.
        if encoding is not None:
            parser = lxml.html.HTMLParser(encoding=encoding)
        else:
            parser = lxml.html.HTMLParser()
        tree = etree.parse(resource_info.filename, parser)
        tree = remove_namespaces(tree)
        return tree

    def extract_value(self, resource_info):
        if not resource_info.content_type in MARKUP_TYPES:
            raise NoValueExtracted

        encoding = self._sniff_encoding(resource_info)

        tree = self._get_tree(resource_info, encoding)
        nodes = tree.xpath(self.xpath)

        if len(nodes) == 0:
            raise NoValueExtracted

        elif len(nodes) > 1:
            log.warn(
                "XPath expression '{}' returned {} results for document at "
                "{}, using only first one".format(
                    self.xpath, len(nodes), resource_info.url_info['loc']))

        target_node = nodes[0]
        text = target_node.text_content()

        if isinstance(text, etree._ElementUnicodeResult):
            text = unicode(text)
        elif isinstance(text, etree._ElementStringResult):
            text = text.decode(encoding)
        else:
            log.error(
                "Unexpected return type for xpath(), expected"
                "'_ElementUnicodeResult'. (Node: {})".format(target_node))
            raise NoValueExtracted

        return text


class DescriptionExtractor(MetadataExtractor):

    def extract_value(self, resource_info):
        value = resource_info.metadata.get('description')
        if value is None:
            raise NoValueExtracted
        return safe_unicode(value)


class CreatorExtractor(MetadataExtractor):

    def extract_value(self, resource_info):
        value = resource_info.metadata.get('creator')
        if value is None:
            raise NoValueExtracted
        return safe_unicode(value)


class SnippetTextExtractor(TextExtractor, MetadataExtractor,
                           HTTPHeaderExtractor):

    def _get_title(self, resource_info):
        extractor = TitleExtractor()
        title = extractor.extract_value(resource_info)
        return title.strip()

    def _get_plain_text(self, resource_info):
        extractor = PlainTextExtractor()
        plain_text = extractor.extract_value(resource_info)
        return plain_text.strip()

    def extract_value(self, resource_info):
        plain_text = safe_unicode(self._get_plain_text(resource_info))
        title = safe_unicode(self._get_title(resource_info))

        snippet_text = plain_text
        # strip title at start of plain text
        if title is not None and snippet_text.startswith(title):
            snippet_text = snippet_text.lstrip(title)
        return safe_unicode(snippet_text)


class LastModifiedExtractor(URLInfoExtractor, HTTPHeaderExtractor):

    def extract_value(self, resource_info):
        if 'lastmod' in resource_info.url_info:
            datestring = resource_info.url_info['lastmod']
            utc_dt = from_iso_datetime(datestring)
            return utc_dt

        if 'last-modified' in resource_info.headers:
            # TODO: We rely on requests.structures.CaseInsensitiveDict here
            utc_dt = from_iso_datetime(resource_info.headers['last-modified'])
            return utc_dt

        utc_dt = IndexingTimeExtractor().extract_value(resource_info)
        return utc_dt


class FilenameExtractor(HTTPHeaderExtractor):

    def extract_value(self, resource_info):
        if 'content-disposition' in resource_info.headers:
            # TODO: We rely on requests.structures.CaseInsensitiveDict here
            header_value = resource_info.headers['content-disposition']
            items = [i.strip() for i in header_value.split(';')]
            for item in items:
                if item.lower().startswith('filename'):
                    key, value = [token.strip() for token in item.split('=')]
                    filename = value.replace('"', '')
                    # TODO: Deal with encoding of non-ASCII filenames
                    return filename.decode('utf-8', errors='replace')
        raise NoValueExtracted


class KeywordsExtractor(MetadataExtractor):

    def extract_value(self, resource_info):
        value = resource_info.metadata.get('keywords')
        if value is None:
            raise NoValueExtracted
        if ',' in value:
            keywords = value.split(',')
        else:
            keywords = value.split()
        return [safe_unicode(kw.strip()) for kw in keywords]


class ConstantExtractor(ResourceIndependentExtractor):

    def __init__(self, value):
        self.value = value

    def extract_value(self, resource_info):
        value = self.value

        if isinstance(value, str):
            value = safe_unicode(value)

        if self.field.multivalued:
            value = [safe_unicode(v) for v in self.value]

        return value


class IndexingTimeExtractor(ResourceIndependentExtractor):

    def extract_value(self, resource_info):
        return datetime.utcnow()


class SiteAttributeExtractor(SiteConfigExtractor):

    def __init__(self, key):
        self.key = key

    def extract_value(self, resource_info):
        value = resource_info.site.attributes.get(self.key)

        if value is None:
            raise NoValueExtracted

        if isinstance(value, str):
            value = safe_unicode(value)

        return value


class HeaderMappingExtractor(HTTPHeaderExtractor):

    def __init__(self, header_name, mapping, default=None):
        self.header_name = header_name
        self.mapping = mapping
        self.default = default

    def _default_or_raise(self):
        if self.default is not None:
            return safe_unicode(self.default)
        else:
            raise NoValueExtracted

    def extract_value(self, resource_info):
        header_value = resource_info.headers.get(self.header_name)
        if header_value is None:
            # Header not present
            return self._default_or_raise()

        if self.header_name.lower() == 'content-type':
            header_value = get_content_type(header_value)

        if header_value in self.mapping:
            return safe_unicode(self.mapping[header_value])
        else:
            # Header present but not mapped
            return self._default_or_raise()


class FieldMappingExtractor(HTTPHeaderExtractor):

    def __init__(self, field_name, mapping, default=None):
        self.field_name = field_name
        self.mapping = mapping
        self.default = default

    def _default_or_raise(self):
        if self.default is not None:
            return safe_unicode(self.default)
        else:
            raise NoValueExtracted

    def extract_value(self, resource_info):
        mapped_field = self.field.config.get_field(self.field_name)
        field_value = mapped_field.extractor.extract_value(resource_info)
        if field_value is None:
            # Field not extracted
            return self._default_or_raise()

        if field_value in self.mapping:
            return safe_unicode(self.mapping[field_value])
        else:
            # Field present but not mapped
            return self._default_or_raise()
